"""Functions for analysing texts and strings"""

from typing import List, Dict, Tuple
import copy

import numpy as np
import pandas as pd
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.metrics.pairwise import cosine_similarity

import Levenshtein as levenshtein
from Levenshtein import distance as lev


def cosine_sim(input_text_list: List[str], stopwords = 'english') -> np.float64:
    
    """
    Calculates the cosine similarity of two texts based on their word frequencies.
    
    Parameters
    ----------
    input_text_list : list
        list of strings to compare.
    stopwords : str
        name of stopwords dataset to use.
    
    Returns
    -------
    result : numpy.float64
        a float representing the count-vectorised cosine similarity of the inputted strings.
    
    Notes
    -----
    Takes a list of two texts.
    Removes stopwords based on the 'stopwords' argument.
    """
    
    # Initialising vectoriser object
    count_vectorizer = CountVectorizer(stop_words=stopwords)
    
    # Vectorising texts based on word frequency counts and transforming to a sparse matrix
    sparse_matrix = count_vectorizer.fit_transform(input_text_list)
    
    # Transforming sparse matrix to dense matrix
    doc_term_matrix = sparse_matrix.todense()
    
    # Creating dataframe from word vectors matrix
    df = pd.DataFrame(
       doc_term_matrix,
       columns = count_vectorizer.get_feature_names_out(),
       index = ['input1', 'input2'],
    )
    
    # Running cosine similarity algorithm on dataframe
    result = cosine_similarity(df, df)
    
    return result[0,1]

def normalised_levenshtein(first_string: str, second_string: str) -> float:
    
    """
    Calculates the normalised levenshtein distance between two strings.
    
    Parameters
    ----------
    first_string : str
        first string to compare.
    second_string : str
        second string to compare.
    
    Returns
    -------
    result : float
        a float representing the normalised Levenshtein similarity of the inputted strings.
    
    Notes
    -----
    Normalisation function: levenshtein distance divided by the length of the longest string.
    """
    
    # Checking types
    if type(first_string) != str:
        raise TypeError('Inputted strings must be of type "str"')
    
    if type(second_string) != str:
        raise TypeError('Inputted strings must be of type "str"')
    
    # Getting length of longest string
    len1 =  len(first_string)
    len2 = len(second_string)
    max_len = max(len1, len2)
    
    # Calculating levenshtein distance
    levenshtein = lev(first_string, second_string)
    
    # Normalising result
    normalised = levenshtein / max_len
    
    return normalised
    

def inverse_normalised_levenshtein(first_string: str, second_string: str) -> float:
    
    """
    Calculates the inverse of the normalised levenshtein distance between two strings.
    
    Parameters
    ----------
    first_string : str
        first string to compare.
    second_string : str
        second string to compare.
    
    Returns
    -------
    result : float
        a float representing the inverse normalised Levenshtein similarity of the inputted strings.
    
    Notes
    -----
    Normalisation function: levenshtein distance divided by the length of the longest string.
    """
    
    # Checking types
    if type(first_string) != str:
        raise TypeError('Inputted strings must be of type str')
    
    if type(second_string) != str:
        raise TypeError('Inputted strings must be of type str')
    
    # Calculating normalised levenshtein distance
    normalised_lev = normalised_levenshtein(first_string = first_string, second_string = second_string)
    
    # Calculating inverse
    inv_normalised = 1 - normalised_lev
    
    return inv_normalised


"""Functions for exporting network objects to files"""

from typing import List, Dict, Tuple
import json
import copy
import igraph as ig # type: ignore
from igraph import Graph # type: ignore
from networkx.classes import Graph as NetworkX_Undir, DiGraph as NetworkX_Dir, MultiGraph as NetworkX_Multi # type: ignore

def export_network_to_kumu(network: Graph, folder_address: str = 'request_input', file_name: str = 'request_input'):
        
        """
        Exports network object (e.g. igraph.Graph, CaseNetwork, NetworkX) as a JSON file to be uploaded to Kumu.
        """
        
        # Requesting folder address from user input
        if folder_address == 'request_input':
            folder_address = input('Save to: ')

        # Requesting file name from user input
        if file_name == 'request_input':
            file_name = input('Save as: ')
        
        # Creating file address
        file_address = folder_address + '/' + file_name + '.json'
        
        # Initialising network blueprint variables
        kumu_blueprint = {}
        elements_list = []
        edge_list = []
        
        # Converting NetworkX objects to igraph objects
        if (
            (type(network) == NetworkX_Undir)
            or (type(network) == NetworkX_Dir)
            or (type(network) == NetworkX_Multi)
        ):
            network = Graph.from_networkx(network)
        
        # Setting 'type' attribute and handling potential errors (e.g. if 'type' has another name)
        if 'type' not in network.vs.attributes():
            network.vs['type'] = network.vs['type_1']
    
            if 'type_1' not in network.vs.attributes():
                network.vs['type'] = None
        
        # Setting network directed attribute
        if network.is_directed() == True:
            direction = "directed"

        else:
            direction = "mutual"
        
        # Creating list of vertices
        for vertex in network.vs():

            element_dict = {}
            element_dict["label"] = vertex["name"]
            element_dict["type"] = vertex["type"]

            elements_list.append(element_dict)

        # Creating edgelist
        for edge in network.es():
            edge_dict = {}
            edge_dict["from"] = network.vs[edge.source]["name"]
            edge_dict["to"] = network.vs[edge.target]["name"]
            edge_dict["direction"] = direction
    #         edge_dict["weight"] = edge["weight"]

            edge_list.append(edge_dict)

        # Creating dictionary
        kumu_blueprint["elements"] = elements_list
        kumu_blueprint["connections"] = edge_list

        
        # Exporting dictionary as JSON
        with open(file_address, "w") as outfile:
            json.dump(kumu_blueprint, outfile)

    

def export_network(network: Graph, file_name: str = 'request_input', folder_address: str = 'request_input', file_type: str = 'request_input'):
        
        """
        Exports network object (e.g. igraph.Graph, CaseNetwork, NetworkX) to one of a variety of graph file types. Defaults to .graphML.
        """
        
        # Requesting folder address from user input
        if folder_address == 'request_input':
            folder_address = input('Folder address: ')
        
        # Requesting file name from user input
        if file_name == 'request_input':
            file_name = input('Save as: ')
        
        # Requesting file type from user input
        if file_type == 'request_input':
            file_type = input('File type ("graphML", "kumu", GML", "LEDA", "lgl", "ncol", or "pajek"): ')
        
        # Converting NetworkX objects to igraph objects
        if (
            (type(network) == NetworkX_Undir)
            or (type(network) == NetworkX_Dir)
            or (type(network) == NetworkX_Multi)
        ):
            network = Graph.from_networkx(network)
        
        # Writing GraphML file by default or if selected
        if (file_type == None) or (file_type == '') or (file_type.lower() == 'graphml'):

                file_address = folder_address + '/' + file_name + '.graphml'
                network.write_graphml(file_address)
        
        # Writing Kumu formatted JSON file if selected
        if (file_type.lower() == 'kumu') or (file_type.lower() == 'kumu blueprint'):
            
            export_network_to_kumu(network = network, folder_address = folder_address, file_name = file_name)
        
        # Writing GML file if selected
        if file_type.lower() == 'gml':

                file_address = folder_address + '/' + file_name + '.gml'
                network.write_gml(file_address)
        
        # Writing LEDA file if selected
        if file_type.lower() == 'leda':

                file_address = folder_address + '/' + file_name + '.lgr'
                network.write_leda(file_address)
        
        # Writing LGL file if selected
        if file_type.lower() == 'lgl':

                file_address = folder_address + '/' + file_name + '.lgl'
                network.write_lgl(file_address)
        
        # Writing NCOL file if selected
        if file_type.lower() == 'ncol':

                file_address = folder_address + '/' + file_name + '.ncol'
                network.write_ncol(file_address)
        
        # Writing Pajek file if selected
        if file_type.lower() == 'pajek':

                file_address = folder_address + '/' + file_name + '.net'
                network.write_pajek(file_address)

                


from ..utils.cleaners import deduplicate
from ..importers.crossref import search_funder_works, lookup_funder
from ..datasets.stopwords.stopwords import all_stopwords

from .entities import Entity, Entities

# from .results import Results

import pandas as pd
import numpy as np



from nltk.tokenize import word_tokenize # type: ignore

def generate_funder_id(funder_data: pd.Series):

        funder_data = funder_data.copy(deep=True).dropna().astype(str).str.lower()

        funder_id = 'F:'

        if 'name' in funder_data.index:
            name = funder_data['name']
        else:
            name = ''

        if (name == None) or (name == ''):
            name = 'no_name_given'
        
        if name != 'no_name_given':

            name = name.strip().lower()
            name_tokens = list(word_tokenize(name))
            name_tokens = [i for i in name_tokens if ((i not in all_stopwords) and (i != "'s"))]
            name_first_3 = name_tokens[:3]
            name_last = name_tokens[-1]

            if name_last in name_first_3:
                name_last = ''
            
            name_shortened = '-'.join(name_first_3) + '-' + name_last

        else:
            name_shortened = name

        funder_id = funder_id + '-' + name_shortened
        
        uid = ''

        if 'uri' in funder_data.index:
            uid = funder_data['uri']

            if (uid == None) or (uid == 'None') or (uid == ''):
                if 'crossref_id' in funder_data.index:
                    uid = funder_data['crossref_id']
                    
                    if (uid == None) or (uid == 'None') or (uid == ''):
                        if 'website' in funder_data.index:
                            uid = funder_data['website']

                            if (uid == None) or (uid == 'None') or (uid == ''):
                                    uid = ''
        
        uid_shortened = uid.replace('https://', '').replace('http://', '').replace('www.', '').replace('dx.','').replace('doi.org/','').replace('user=','')[:24]

        funder_id = funder_id + '-' + uid_shortened
        funder_id = funder_id.replace('F:-', 'F:').replace("'s", '').replace('\r', '').replace('\n', '').replace('.', '').replace("'", "").replace('"', '').replace('(','').replace(')','').replace('`','').replace('’','').replace('--', '-').replace('F:-', 'F:').strip('-')

        return funder_id

class Funder(Entity):

    """
    This is a Funder object. It is designed to store data about an individual funder and their publications.
    
    Parameters
    ----------
    
    
    Attributes
    ----------
    """

    def __init__(self,
                 funder_id: str = None, # type: ignore
                 name: str = None, # type: ignore
                 alt_names = [], # type: ignore
                 location: str = None, # type: ignore
                 email: str = None, # type: ignore
                 uri: str = None, # type: ignore
                 crossref_id: int = None, # type: ignore
                 work_count: int = None,  # type: ignore
                 tokens = [], # type: ignore
                 website: str = None,  # type: ignore
                 other_links = [], # type: ignore
                 use_api: bool = False
                 ):
        
        """
        Initialises funder instance.
        
        Parameters
        ----------
        """

        super().__init__()

        if type(name) == str:
            name = name.strip()

        if type(alt_names) == str:
            alt_names = alt_names.replace('[','').replace(']','').split(',')
            alt_names = [i.strip() for i in alt_names if type(i) == str]
        
        if name == None:
            if type(alt_names) == str:
                name = alt_names
            
            else:
                if (type(alt_names) == list) and (len(alt_names) > 0) and (type(alt_names[0]) == str):
                    name = alt_names[0]
        
        if type(other_links) == str:
            other_links = other_links.replace('[','').replace(']','').split(',')
            other_links = [i.strip() for i in other_links]
        
        if type(tokens) == str:
            tokens = tokens.replace('[','').replace(']','').split(',')
            tokens = [i.strip() for i in tokens]
        

        self.summary = pd.DataFrame(columns = [
                                'funder_id',
                                'name',
                                'alt_names',
                                'location',
                                'email',
                                'uri',
                                'crossref_id',
                                'work_count',
                                'tokens',
                                'website',
                                'other_links'
                                ],
                                dtype = object)
        
        
        self.summary.loc[0] = pd.Series(dtype=object)
        self.summary.loc[0, 'funder_id'] = funder_id
        self.summary.loc[0, 'name'] = name
        self.summary.at[0, 'alt_names'] = alt_names
        self.summary.loc[0, 'location'] = location
        self.summary.loc[0, 'email'] = email
        self.summary.loc[0, 'uri'] = uri
        self.summary.loc[0, 'crossref_id'] = crossref_id
        self.summary.loc[0, 'work_count'] = work_count   
        self.summary.at[0, 'tokens'] = tokens
        self.summary.loc[0, 'website'] = website
        self.summary.at[0, 'other_links'] = other_links

        # self.publications = Results()

        if use_api == True:
            self.update_from_crossref()
        
        self.update_id()

    def generate_id(self):

        funder_data = self.summary.loc[0]

        funder_id = generate_funder_id(funder_data) # type: ignore
        return funder_id

    def update_id(self):

        current_id = self.summary.loc[0, 'funder_id']

        if (current_id == None) or (current_id == 'None') or (current_id == '') or (current_id == 'F:000'):
            auth_id = self.generate_id()
            self.summary.loc[0, 'funder_id'] = auth_id
        
    
    def __getitem__(self, key):
        
        """
        Retrieves funder attribute using a key.
        """
        
        if key in self.__dict__.keys():
            return self.__dict__[key]
        
        if key in self.summary.columns:
            return self.summary.loc[0, key]
        
        # if key in self.publications.columns:
        #     return self.publications[key]

    def __repr__(self) -> str:
        
        return str(self.summary.loc[0, 'name'])

    def has_uri(self) -> bool:

        uri = self.summary.loc[0, 'uri']

        if (type(uri) == str) and (uri != ''):
            return True
        else:
            return False

    def add_dict(self, data: dict):

        if 'name' in data.keys():
            name = data['name']
            self.summary.loc[0, 'name'] = name

        if 'doi' in data.keys():
            uri = data['doi'].replace('http', '').replace('https', '').replace('dx.', '').replace('doi.org/', '').strip()
            self.summary.loc[0, 'uri'] = 'https://doi.org/' + uri
    
    def from_dict(data: dict, use_api=False): # type: ignore

        funder = Funder()
        funder.add_dict(data=data)

        if use_api == True:
            funder.update_from_crossref()

        return funder
        
    def add_series(self, series: pd.Series):
        self.summary.loc[0] = series

    def from_series(data: pd.Series): # type: ignore
        funder = Funder()
        funder.add_series(data)

        return funder

    def add_dataframe(self, dataframe: pd.DataFrame):
        series = dataframe.loc[0]
        self.add_series(series) # type: ignore

    def from_dataframe(data: pd.DataFrame): # type: ignore
        funder = Funder()
        funder.add_dataframe(data)

        return funder

    def import_crossref_result(self, crossref_result: pd.Series):
        
        if 'name' in crossref_result.index:
            name = crossref_result['name']
        else:
            name = self.summary.loc[0, 'name']

        if 'alt-names' in crossref_result.index:
            alt_names = crossref_result['alt-names']
        else:
            alt_names = self.summary.loc[0, 'alt_names']

        if 'location' in crossref_result.index:
            location = crossref_result['location']
        else:
            location = self.summary.loc[0, 'location']

        if 'email' in crossref_result.index:
            email = crossref_result['email']
        else:
            email = self.summary.loc[0, 'email']

        if 'uri' in crossref_result.index:
            uri  =crossref_result['uri']
        else:
            uri = self.summary.loc[0, 'uri']

        if 'id' in crossref_result.index:
            crossref_id = crossref_result['id']
        else:
            crossref_id = self.summary.loc[0, 'crossref_id']

        if 'work-count' in crossref_result.index:
            work_count = crossref_result['work-count']
        else:
            work_count = self.summary.loc[0, 'work_count']

        if 'tokens' in crossref_result.index:
            tokens = crossref_result['tokens']
        else:
            tokens = self.summary.loc[0, 'tokens']
        
        self.summary.loc[0, 'name'] = name
        self.summary.loc[0, 'alt_names'] = alt_names
        self.summary.loc[0, 'location'] = location
        self.summary.loc[0, 'email'] = email
        self.summary.loc[0, 'uri'] = uri
        self.summary.loc[0, 'crossref_id'] = crossref_id
        self.summary.loc[0, 'work_count'] = work_count
        self.summary.loc[0, 'tokens'] = tokens
    
    def from_crossref_result(crossref_result: pd.Series): # type: ignore

        funder = Funder()
        funder.import_crossref_result(crossref_result=crossref_result)

        return funder

    def import_crossref(self, crossref_id: str, timeout = 60):

        res = lookup_funder(crossref_id, timeout)
        self.import_crossref_result(res.loc[0]) # type: ignore

    def from_crossref(crossref_id: str): # type: ignore
        
        funder = Funder()
        funder.import_crossref(crossref_id) # type: ignore

        return funder
    
    def import_uri(self, uri: str, timeout = 60):

        res = lookup_funder(uri, timeout)
        self.import_crossref_result(res.loc[0]) # type: ignore

    def from_uri(uri: str): # type: ignore
        
        funder = Funder()
        funder.import_uri(uri) # type: ignore

        return funder

    def update_from_crossref(self, timeout = 60):

        uid = self.summary.loc[0,'crossref_id']
        if uid == None:
            uid = self.summary.loc[0,'uri']
            if uid == None:
                uid = ''

        res = lookup_funder(funder_id = uid, timeout = timeout) # type: ignore
        if len(res) > 0:
            self.import_crossref_result(res.loc[0]) # type: ignore

    def update_from_uri(self, timeout = 60):

        uid = self.summary.loc[0,'uri']
        if uid == None:
            uid = self.summary.loc[0,'crossref']
            if uid == None:
                uid = ''

        res = lookup_funder(funder_id = uid, timeout = timeout) # type: ignore
        if len(res) > 0:
            self.import_crossref_result(res.loc[0]) # type: ignore
    
    def search_works(self, bibliographic: str = None, # type: ignore
                     title: str = None, # type: ignore
                     author: str = None, # type: ignore
                     author_affiliation: str = None, # type: ignore
                     editor: str = None, # type: ignore
                     entry_type: str = None, # type: ignore
                     published_date: str = None, # type: ignore
                     doi: str = None, # type: ignore
                     publisher_name: str = None,# type: ignore
                    source: str = None, # type: ignore
                    link: str = None, # type: ignore
                    filter: dict = None, # type: ignore
                    select: list = None, # type: ignore
                    sample: int = None, # type: ignore
                    limit: int = 10, 
                    rate_limit: float = 0.05, 
                    timeout: int = 60, 
                    add_to_publications = False) -> pd.DataFrame:
        
        uid = self.summary.loc[0, 'crossref_id']
        if (uid == None) or (uid == ''):
            uid = self.summary.loc[0, 'uri']
            if (uid == None) or (uid == ''):
                uid = ''
        
        uid = str(uid)


        result = search_funder_works(funder_id=uid,  
                                     bibliographic=bibliographic, 
                                     title=title, author=author, 
                                     author_affiliation=author_affiliation, 
                                     editor=editor,
                                     entry_type=entry_type,
                                     published_date=published_date,
                                     doi=doi,
                                     publisher_name=publisher_name,
                                     source=source,
                                     link=link,
                                     filter=filter,
                                     select=select,
                                     sample=sample,
                                     limit=limit,
                                     rate_limit=rate_limit,
                                     timeout=timeout
                                     )
        
        # if add_to_publications == True:
        #     self.publications.add_dataframe(result)
        
        return result

class Funders(Entities):

    """
    This is a Funders object. It contains a collection of Funders objects and compiles data about them.
    
    Parameters
    ----------
    
    
    Attributes
    ----------
    """

    def __init__(self, funders_data = None):
        
        """
        Initialises Funders instance.
        
        Parameters
        ----------
        """

        super().__init__()

        self.summary = pd.DataFrame(columns = 
                                ['funder_id',
                                'name',
                                'alt_names',
                                'location',
                                'email',
                                'uri',
                                'crossref_id',
                                'work_count',
                                'tokens',
                                'website',
                                'other_links'
                                ],
                                dtype = object)
        

        self.all = dict()

        self.data = []
        self.data.append(funders_data)

        if (type(funders_data) == list) and (type(funders_data[0]) == Funder):

            for i in funders_data:
                fu = i.summary.copy(deep=True)
                self.summary = pd.concat([self.summary, fu])

            self.summary = self.summary.reset_index().drop('index',axis=1)

        else:

            if type(funders_data) == dict:
                
                values = list(funders_data.values())

                if type(values[0]) == Funder:

                    for f in funders_data.keys():
                        
                        index = len(self.summary)
                        fu = f.summary.copy(deep=True)
                        self.summary = pd.concat([self.summary, fu])
                        self.summary.loc[index, 'funder_id'] = f

                    self.summary = self.summary.reset_index().drop('index',axis=1)
                
        self.update_ids()

    def __getitem__(self, key):
        
        """
        Retrieves funders attribute using a key.
        """
        
        if key in self.__dict__.keys():
            return self.__dict__[key]
        
        if key in self.all.keys():
            return self.all[key]

        if key in self.summary.columns:
            return self.summary[key]
        
        if (type(key) == int) and (key <= len(self.data)):
            return self.data[key]
    
    def __repr__(self) -> str:

        alphabetical = str(self.summary['name'].sort_values().to_list()).replace('[','').replace(']','')
        return alphabetical
    
    def __len__(self) -> int:
        return len(self.all.keys())

    def merge(self, funders, drop_empty_rows = True, drop_duplicates = False):

        left = self.summary.copy(deep=True)
        right = funders.summary.copy(deep=True)
        
        merged = pd.concat([left, right])

        self.summary = merged.drop_duplicates(subset=['funder_id', 'name', 'crossref_id'], ignore_index=True)

        for i in funders.all.keys():
            if i not in self.all.keys():
                self.all[i] = funders.all[i]

        left_data = self.data
        right_data = funders.data

        if left_data == None:
                left_data = []
            
        if right_data == None:
                right_data = []

        if (type(left_data) == Funder) or (type(left_data) == str):
                left_data = [left_data]
            
        if (type(right_data) == Funder) or (type(right_data) == str):
                right_data = [right_data]
            
        if type(left_data) == dict:
                left_data = list(left_data.values())
            
        if type(right_data) == dict:
                right_data = list(right_data.values())

        merged_data = left_data + right_data # type: ignore
        merged_data = pd.Series(merged_data).value_counts().index.to_list()

        self.data = merged_data

        if drop_empty_rows == True:
            self.drop_empty_rows()
        
        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        self.update_ids()

        return self


    def add_funder(self, funder: Funder =  None, uri: str = None, crossref_id: int = None, data = None, use_api = True, drop_empty_rows = False, drop_duplicates = False): # type: ignore

        

        if type(funder) == str:

            if use_api == True:
                details = lookup_funder(funder_id=funder)
                crossref_id = details.loc[0, 'id'] # type: ignore
                uri = details.loc[0, 'uri'] # type: ignore

            else:
                crossref_id = funder

            funder = None # type: ignore

        if data is not None:

            if type(data) == dict:
                funder = Funder.from_dict(data=data, use_api = use_api) # type: ignore
            
            else:
                if type(data) == pd.Series:
                    funder = Funder.from_series(data=data) # type: ignore
                
                else:
                    if type(data) == pd.DataFrame:
                        funder = Funder.from_dataframe(data=data) # type: ignore


        if funder is None:
            funder = Funder(uri=uri, crossref_id=crossref_id)

        if use_api == True:
            funder.update_from_crossref()

        funder.update_id()

        funder_id = str(funder.summary.loc[0, 'funder_id'])

        # if funder_id in self.summary['funder_id'].to_list():
        #     id_count = len(self.summary[self.summary['funder_id'].str.contains(funder_id)]) # type: ignore
        #     funder_id = funder_id + f'#{id_count + 1}'
        #     funder.summary.loc[0, 'funder_id'] = funder_id

        self.summary = pd.concat([self.summary, funder.summary])
        self.summary = self.summary.reset_index().drop('index', axis=1)

        self.all[funder_id] = funder

        if data is None:
            data = funder.summary.to_dict(orient='index')
        
        self.data.append(data)

        self.update_ids()

        if drop_empty_rows == True:
            self.drop_empty_rows()
        
        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)


    def add_funders_list(self, funders_list: list, drop_empty_rows = False, drop_duplicates = False):
        
        for i in funders_list:
            if type(i) == Funder:
                self.add_funder(funder = i)
        
        if drop_empty_rows == True:
            self.drop_empty_rows()
        
        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

    def drop_empty_rows(self):

        ignore_cols = ['funder_id', 'alt_names', 'publications', 'tokens', 'other_links']

        df = self.summary.copy(deep=True)
        df['name'] = df['name'].replace('no_name_given', None)
        df = df.dropna(axis=0, how='all')
        drop_cols = [c for c in df.columns if c not in ignore_cols]
        df = df.dropna(axis=0, how='all', subset=drop_cols).reset_index().drop('index', axis=1)

        self.summary = df

        return self

    def remove_duplicates(self, drop_empty_rows = True, sync = False):

        if drop_empty_rows == True:
            self.drop_empty_rows()
        
        df = self.summary.copy(deep=True)
        df['uri'] = df['uri'].str.replace('http://', '', regex=False).str.replace('https://', '', regex=False).str.replace('wwww.', '', regex=False).str.replace('dx.', '', regex=False).str.replace('doi.org/', '', regex=False).str.strip('/')
        
        df = df.sort_values(by = ['uri', 'crossref_id', 'website', 'name']).reset_index().drop('index', axis=1)
        self.summary = deduplicate(self.summary)

        if sync == True:
            self.sync_details(drop_duplicates=False, drop_empty_rows=False)

        return self

    def sync_all(self, drop_duplicates = False, drop_empty_rows=False):

        for i in self.all.keys():
            funder = self.all[i]
            funder.update_id()
            series = funder.summary.copy(deep=True).loc[0]
            all = self.summary.copy(deep=True).astype(str)
            indexes = all[all['funder_id'] == i].index.to_list()
            if len(indexes) > 0:
                auth_index = indexes[0]
                all_copy = self.summary.copy(deep=True)
                all_copy.loc[auth_index] = series
                self.summary = all_copy
        
        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

    def sync_details(self, drop_duplicates = False, drop_empty_rows=False):

        self.update_ids(sync=False)

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        for i in self.summary.index:

            f_data = self.summary.loc[i]
            f_id = f_data['funder_id']

            if f_id != None:
                f = Funder.from_series(f_data) # type: ignore
                self.all[f_id] = f

            else:
                f_id = generate_funder_id(f_data)
                f_data['funder_id'] = f_id
                f = Funder.from_series(f_data) # type: ignore
                self.all[f_id] = f
        
        keys = list(self.all.keys())
        for key in keys:
            f_ids = self.summary['funder_id'].to_list()
            if key not in f_ids:
                del self.all[key]

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)


    def sync(self, drop_duplicates = False, drop_empty_rows=False):
        
        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        all_len = len(self.summary)
        details_len = len(self.all)

        if all_len > details_len:
            self.sync_details(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
            return
        else:
            if details_len > all_len:
                self.sync_all(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
                return
            else:
                self.sync_details(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
                self.sync_all(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
                return

    def update_ids(self, sync=False):
        
        if sync == True:
            self.sync()

        for i in self.summary.index:
            all_copy = self.summary.copy(deep=True)
            data = all_copy.loc[i]
            old_id = all_copy.loc[i, 'funder_id']
            new_id = generate_funder_id(data)

            # if new_id in self.summary['funder_id'].to_list():
            #     df_copy = self.summary.copy(deep=True)
            #     df_copy = df_copy.astype(str)
            #     id_count = len(df_copy[df_copy['funder_id'].str.contains(new_id)]) # type: ignore
            #     new_id = new_id + f'#{id_count + 1}'

            all_copy2 = self.summary.copy(deep=True)
            all_copy2.loc[i, 'funder_id'] = new_id
            self.summary = all_copy2

            if old_id in self.all.keys():
                self.all[new_id] = self.all[old_id]
                self.all[new_id].summary.loc[0, 'funder_id'] = new_id
                del self.all[old_id]

            else:
                funder = Funder.from_series(data) # type: ignore
                funder.summary.loc[0, 'funder_id'] = new_id
                self.all[new_id] = funder


    def update_from_crossref(self, drop_duplicates = False, drop_empty_rows=False):

        funder_ids = self.all.keys()

        for a in funder_ids:

            self.all[a].update_from_crossref()
            details = self.all[a].summary.loc[0]
            
            df_index = self.summary[self.summary['funder_id'] == a].index.to_list()[0]
            self.summary.loc[df_index] = details

            new_id = details['funder_id']
            if new_id != a:
                self.all[new_id] = self.all[a]
                del self.all[a]

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        self.update_ids()
        


    def import_crossref_ids(self, crossref_ids: list, drop_duplicates = False, drop_empty_rows=False):

        for i in crossref_ids:

            auth = Funder.from_crossref(i) # type: ignore
            self.add_funder(funder = auth, data = i)

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        self.update_ids()


    def from_crossref_ids(crossref_ids: list, drop_duplicates = False, drop_empty_rows=False): # type: ignore

        funders = Funders()
        funders.import_crossref_ids(crossref_ids, drop_duplicates = drop_duplicates, drop_empty_rows=drop_empty_rows)

        return funders

    def with_crossref(self):
        return self.summary[~self.summary['crossref_id'].isna()]
    
    def with_uri(self):
        return self.summary[~self.summary['uri'].isna()]

    def import_crossref_result(self, crossref_result: pd.DataFrame, use_api = False, drop_duplicates = False, drop_empty_rows=False):

        for i in crossref_result.index:
            
            data = crossref_result.loc[i]
            fu = Funder.from_crossref_result(crossref_result=data) # type: ignore
            self.add_funder(funder = fu, data = data, use_api=use_api)
        
        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        self.update_ids()


    def from_crossref_result(crossref_result: pd.DataFrame, use_api = False, drop_duplicates = False, drop_empty_rows=False): # type: ignore

        funders = Funders()
        funders.import_crossref_result(crossref_result, use_api=use_api, drop_duplicates = drop_duplicates, drop_empty_rows=drop_empty_rows)

        return funders

    def search_works(self, 
                     funder_id: str = None, # type: ignore
                    index: int = None, # type: ignore
                    crossref_id: int = None, # type: ignore
                    uri: str = None,# type: ignore
                     bibliographic: str = None, # type: ignore
                     title: str = None, # type: ignore
                     author: str = None, # type: ignore
                     author_affiliation: str = None, # type: ignore
                     editor: str = None, # type: ignore
                     entry_type: str = None, # type: ignore
                     published_date: str = None, # type: ignore
                     doi: str = None, # type: ignore
                     publisher_name: str = None,# type: ignore
                    source: str = None, # type: ignore
                    link: str = None, # type: ignore
                    filter: dict = None, # type: ignore
                    select: list = None, # type: ignore
                    sample: int = None, # type: ignore
                    limit: int = 10, 
                    rate_limit: float = 0.05, 
                    timeout: int = 60, 
                    add_to_publications = False) -> pd.DataFrame:
        
        if (funder_id != None) and (funder_id in self.all.keys()):
            funder = self.all[funder_id]
            result = funder.search_works(
                                     bibliographic=bibliographic, 
                                     title=title, author=author, 
                                     author_affiliation=author_affiliation, 
                                     editor=editor,
                                     entry_type=entry_type,
                                     published_date=published_date,
                                     doi=doi,
                                     publisher_name=publisher_name,
                                     source=source,
                                     link=link,
                                     filter=filter,
                                     select=select,
                                     sample=sample,
                                     limit=limit,
                                     rate_limit=rate_limit,
                                     timeout=timeout,
                                     add_to_publications=add_to_publications
                                     )
        
        else:
            if index != None:
    
                uid = self.summary.loc[index, 'crossref_id']
                if (uid==None) or (uid == ''):
                    uid = self.summary.loc[index, 'uri']
                    if (uid==None) or (uid == ''):
                        uid = ''

                funder_id = self.summary.loc[index, 'funder_id'] # type: ignore

                uid = str(uid)
            
            else:
                if crossref_id != None:
                    index = self.summary[self.summary['crossref_id'] == crossref_id].index.to_list()[0]
                    funder_id = self.summary.loc[index, 'funder_id'] # type: ignore
                    uid = str(crossref_id)
                    
                else:
                    if uri != None:
                        index = self.summary[self.summary['uri'] == uri].index.to_list()[0]
                        funder_id = self.summary.loc[index, 'funder_id'] # type: ignore
                        uid = str(uri)
                        

            result = search_funder_works(funder_id=uid,  
                                     bibliographic=bibliographic, 
                                     title=title, author=author, 
                                     author_affiliation=author_affiliation, 
                                     editor=editor,
                                     entry_type=entry_type,
                                     published_date=published_date,
                                     doi=doi,
                                     publisher_name=publisher_name,
                                     source=source,
                                     link=link,
                                     filter=filter,
                                     select=select,
                                     sample=sample,
                                     limit=limit,
                                     rate_limit=rate_limit,
                                     timeout=timeout
                                     )
        
            if add_to_publications == True:
                self.all[funder_id].publications.add_dataframe(result)
        
        return result

def format_funders(funder_data, use_api = False, drop_duplicates = False, drop_empty_rows=False):
        
        result = Funders()

        funder_type = type(funder_data)

        if (
            (funder_type != pd.DataFrame) 
            and (funder_type != pd.Series) 
            and (funder_type != Funder) 
            and (funder_type != Funders) 
            and (funder_type != list) 
            and (funder_type != dict)
            ):
            result = Funders()
            return result

        if funder_type == Funders:
            result = funder_data
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result

        if funder_type == Funder:
            result.add_funder(funder=funder_data, use_api=use_api)
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result

        if funder_type == pd.Series:
            funder = Funder()
            funder.add_series(funder_data)
            result.add_funder(funder=funder, use_api=use_api)
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result

        if funder_type == pd.DataFrame:
            result.import_crossref_result(funder_data, use_api=use_api) # type: ignore
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result
        
        if (funder_type == dict):
            funder = Funder.from_dict(funder_data) # type: ignore
            result.add_funder(funder = funder, use_api=use_api) # type: ignore
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result

        if (funder_type == list) and (len(funder_data) > 0) and (type(funder_data[0]) == Funder):
            result = Funders()
            result.add_funders_list(funder_data)
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result

        if (funder_type == list) and (len(funder_data) > 0) and (type(funder_data[0]) == dict):

            for i in funder_data:
                funder = Funder.from_dict(i) # type: ignore
                result.add_funder(funder = funder, use_api=use_api) # type: ignore

            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)

            return result


"""Functions for url, domain, website, IP and WhoIs analysis"""

from typing import List
import copy
import re
import ipaddress
import socket
import webbrowser
import numpy as np
import pandas as pd
from courlan import check_url
from urllib.parse import quote


def is_url(url: str) -> bool:
    
    url_regex = re.compile(r'https?://(?:www\.)?[a-zA-Z0-9./]+')
    return bool(url_regex.match(url))

def domain_splitter(web_address: str) -> str:
    
    """
    Takes URL and returns its domain.
    """
    
    # Checking type; converting lists, sets, and tuples to strings
    if (
        ((type(web_address) == list)
        or (type(web_address) == set)
        or (type(web_address) == tuple)) 
        and (len(web_address) > 0)
        ):
            web_address = list(web_address)[0]
    
    if type(web_address) != str:
        raise TypeError('Domain_splitter requires a string')
    
    # Removing URL prefixes
    web_address = web_address.replace('https://', '').replace('https=://', '').replace('www.', '')
    
    # Splitting URL into substrings
    split = web_address.split('/')
    
    # Taking the first substring; removing unwanted leading and trailing characters
    domain = split[0].strip('/').strip('.').strip()
    
    return domain


def is_domain(domain_name: str = 'request_input') -> bool:
    
    """
    Checks if string is a valid domain format.
    """
    
    # Requesting domain name from user input if none given
    if domain_name == 'request_input':
        domain_name = input('Domain name to check: ')
    
    # If list is inputted, converts to string
    if (type(domain_name) == list) and (len(domain_name) > 0):
        domain_name = domain_name[0]
    
    # If set or tuple is inputted, converts to string
    if (((type(domain_name) == set)
        or (type(domain_name) == tuple))
        and (len(domain_name) > 0)):
        domain_name = list(domain_name)[0]
    
    # Raising errors if input is not string, list, or set
    if type(domain_name) != str:
        raise TypeError('Domain name to check must be a string')
    
    # Checking the overall length of the domain name; should not exceed 253 characters
    if len(domain_name) > 253:
        return False

    # Regular expression pattern to validate domain name:
    #   - (?!-) ensures that the label doesn't start with a hyphen
    #   - [A-Za-z0-9-]{1,63} ensures that the label contains 1 to 63 alphanumeric characters or hyphens
    #   - (?<!-) ensures that the label doesn't end with a hyphen
    #   - \. matches a dot between labels
    #   - [A-Za-z]{2,} matches a two or more character top-level domain (TLD)
    pattern = r'^((?!-)[A-Za-z0-9-]{1,63}(?<!-)\.)+[A-Za-z]{2,}$'

    # Checking if the domain name matches the pattern
    try:
        if re.match(pattern, domain_name):
            return True
        else:
            return False
    except:
        return False


def is_ip_address(string: str = 'request_input') -> bool:
    
    """
    Checks if string is a valid IP address.
    """
    
    # Requesting string from user input if none given
    if string == 'request_input':
        string = input('String to check: ')
    
    # Converting lists, sets, and tuples to strings; else raising TypeError
    if ((type(string) == list)
        or (type(string) == set)
        or (type(string) == tuple)
        ) and (len(string) > 0):
        string = list(string)[0]
    
    if type(string) != str:
        raise TypeError('IP address to check must be a string')
    
    # Cleaning string
    string = string.strip().strip('/').strip('\\').strip('#').strip('-').strip()
    
    # Checking if string is an IP address
    try:
        ipaddress.ip_address(string)
        return True
    
    except:
        return False

def is_registered_domain(domain_name: str) -> bool:
    
    """
    Returns a boolean indicating whether a given domain name is registered.
    """
    
    try:
        w = whois.whois(domain_name)
    except Exception:
        return False
    else:
        return bool(w.domain_name)

def correct_url(url: str) -> str:
    
    """
    If URL lacks an HTTPS prefix, adds one.
    """
    
    if (url.startswith('https://') == False) and (url.startswith('http://') == False):
        url = 'https://' + url
    
    return url

def get_domain(url: str) -> str:
    
    """
    Returns a URL's domain.
    """
    
    # Using check_url() to extract domain
    try:
        domain = check_url(url)[1]
        
        # If check_url() does not find a domain, falls back to domain_splitter()
        if domain == None:
            domain = domain_splitter(url)
    
    # If check_url() raises an error, falls back to domain_splitter()
    except:
        domain = domain_splitter(url)
            
    return domain

def get_my_ip() -> str:
    
    """
    Returns user's IP address.
    """
    
    # Trying to retrieve IP address using socket's gethostbyname() method
    try:
        hostname=socket.gethostname()
        IPAddr=socket.gethostbyname(hostname)
    
    
    except:
        
        # If socket's gethostbyname() method raises an error, trying to use socket's getfqdn() method
        try:
            IPAddr = socket.gethostbyname(socket.getfqdn()) 
            
        except:
            
            # If socket's getfqdn() method raises an error, trying to use socket's connect() and getsockname() methods
            
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            IPAddr = s.getsockname()[0]
            s.close()
    
    return IPAddr

def get_my_ip_geocode():
    
    """
    Returns the geocode associated with user's IP address.
    """
    
    return geocoder.ip('me')

def get_ip_geocode(ip_address: str = 'request_input'):
    
    """
    Returns the geocode associated with an IP address.
    """
    
    # Requesting IP address from user input if none given
    if ip_address == 'request_input':
        ip_address = input('IP address: ')

    return geocoder.ip(ip_address)

def get_ip_coordinates(ip_address: str = 'request_input') -> str:
    
    """
    Returns the coordinates associated with an IP address.
    """
    
    # Requesting IP address from user input if none given
    if ip_address == 'request_input':
        ip_address = input('IP address: ')
    
    # Retrieving geocode
    g = geocoder.ip(ip_address)
    
    # Retrieving coordinates
    coords = str(g.latlng)
    
    return coords

def get_my_ip_coordinates() -> str:
    
    """
    Returns the coordinates associated with user's IP address.
    """
    
    return get_ip_coordinates(ip_address = 'me')

def get_ip_physical_location(ip_address: str = 'request_input') -> str:

    """
    Returns the address associated with an IP address.
    """
    
    # Requesting IP address from user input if none given
    if ip_address == 'request_input':
        ip_address = input('IP address: ')
    
    # Retrieving address associated with IP's coordinates using Geopy
    address = str(geocoder.ip(ip_address).address)
    
    return address

def get_my_ip_physical_location() -> str:
    
    """
    Returns the address associated with user's IP address.
    """
    
    # Retrieving geocode associated with user's IP
    location = get_my_ip_geocode()
    
    # Retrieving address
    address = str(location.address)
    
    return address

def lookup_ip_coordinates(ip_address: str = 'request_input', site: str = 'Google Maps'):
    
    """
    Searches for coordinates associated with an IP address on a chosen mapping platform.
    
    Parameters
    ----------
    ip_address : str
        IP address to look up.
    site : str
        name of mapping platform to use. Defaults to 'Google Maps'.
    """
    
    # Requesting IP address from user input if none given
    if ip_address == 'request_input':
        ip_address = input('IP address: ')
    
    # Retrieving coordinates
    coordinates = get_ip_coordinates(ip_address = ip_address)
    latitude = coordinates[0]
    longitude = coordinates[1]
    
    # Searching for coordinates on mapping platform
    try:
        return lookup_coordinates(latitude = latitude, longitude = longitude, site = site)
    except:
        raise ValueError('Lookup failed. Please check the IP address provided.')


def domain_from_ip(ip_address: str) -> pd.Series:
    
    """
    Returns domain address associated with IP address.
    """
    
    # Cleaning IP address
    ip_address = ip_address.strip().strip('/').strip()
    
    # Checking if string is a valid IP address
    if is_ip_address(ip_address) == True:
        
        # Trying to retrieve domain
        try:
            result = socket.gethostbyaddr(ip_address)
        except socket.herror:
            return "No domain details found"
        
        # Creating Pandas series for result
        output_series = pd.Series(result, index = ['domain_name', 'aliases', 'ip_address'])

        return output_series
    
    else:
        raise ValueError('Address given is not a valid IP address')

def ip_from_domain(domain: str) -> str:
    
    """
    Returns IP address associated with domain address.
    """
    
    # Checking if string is a valid domain address
    if is_domain(domain) == False:
        try:
            domain = domain_splitter(domain)
        except:
            pass
                
        if is_domain(domain) == False:
            domain = None
    
    # Trying to retrieve IP address associated with domain
    try:
        result = str(socket.gethostbyname(domain))
        
    except socket.herror:
        return "No domain details found"
    
    return result


            
            

class WhoisResult:
    
    """
    This is a class to store WhoIs result data.
    
    Parameters
    ----------
    domain : str
        domain to run WhoIs search on.
    ip_address : str
        IP address to run WhoIs search on.
    """
    
    def __init__(self, domain: str = None, ip_address: str = None):
        
        """
        Initialises WhoisResult object.
        
        Parameters
        ----------
        domain : str
            domain to run WhoIs search on.
        ip_address : str
            IP address to run WhoIs search on.
        """
        
        # Creating results dataframe and assigning as attribute
        self.all_results = pd.DataFrame(columns = ['Metadata'], dtype = object)
        self.all_results.index.name = 'Category'
        
        # Cleaning domain if given
        if domain != None:
            if is_domain(domain) == False:
                try:
                    domain = domain_splitter(domain)
                except:
                    pass
                
                if is_domain(domain) == False:
                    domain = None
        
        # Cleaning IP address if given
        if ip_address != None:
            ip_address = ip_address.strip().strip('/').strip()
            if is_ip_address(ip_address) == False:
                ip_address = None
        
        # If a domain is given but no IP address, tries to retrieve associated IP address
        if (domain != None) and (ip_address == None):
            try:
                ip_address = ip_from_domain(domain)
            except:
                ip_address = None
        
        # If an IP address is given but no domain, tries to retrieve associated domain
        if (domain == None) and (ip_address != None):
            try:
                domain = domain_from_ip(ip_address)['domain_name']
            except:
                domain = None
        
        # Assigning domain and IP address as attribute
        self.domain = domain
        self.ip_address = ip_address
        
        # Running domain WhoIs lookup
        try:
            self.domain_whois(domain = domain, ip_address = ip_address)
        except:
            pass
        
        # Running IP WhoIs lookup
        try:
            self.ip_whois(domain = domain, ip_address = ip_address)
        except:
            pass
        
        # If IP WhoIs search has been successful, assigns RDAP result as an attribute
        try:
            self.RDAP_obj = self.IPWhois_obj.lookup_rdap(depth=1)
        except:
            self.RDAP_obj = None
        
        # If IP WhoIs search has been successful, assigns RDAP result as an attribute
        try:
            self.results_dict = self.RDAP_obj
        except:
            self.results_dict = None
        
        # If WhoIs search has been successful, assigns rdap_res result to results dataframe
        try:
            rdap_res = pd.Series(self.RDAP_obj)
            df = pd.DataFrame(rdap_res, columns = ['Metadata'])
            df.index.name = 'Category'
            self.all_results = pd.concat([df, self.all_results])
        except:
            pass
        
        # If WhoIs search has been successful, assigns nir result as an attribute
        try:
            self.nir = self.all_results.loc['nir', 'Metadata']
        except:
            self.nir = None
        
        # If WhoIs search has been successful, assigns asn result as an attribute
        try:
            self.asn_registry = self.all_results.loc['asn_registry', 'Metadata']
        except:
            self.asn_registry = None
        
        # If WhoIs search has been successful, assigns asn result as an attribute
        try:
            self.asn = self.all_results.loc['asn', 'Metadata']
        except:
            self.asn = None
        
        # If WhoIs search has been successful, assigns asn_cidr result as an attribute
        try:
            self.asn_cidr = self.all_results.loc['asn_cidr', 'Metadata']
        except:
            self.asn_cidr = None
        
        # If WhoIs search has been successful, assigns asn country code result as an attribute
        try:
            self.asn_country_code = self.all_results.loc['asn_country_code', 'Metadata']
        except:
            self.asn_country_code = None
        
         # If WhoIs search has been successful, assigns asn date country code result as an attribute
        try:
            self.asn_date = self.all_results.loc['asn_date', 'Metadata']
        except:
            self.asn_date = None
        
         # If WhoIs search has been successful, assigns asn description code result as an attribute
        try:
            self.asn_description = self.all_results.loc['asn_description', 'Metadata']
        except:
            self.asn_description = None
        
         # If WhoIs search has been successful, assigns query result as an attribute
        try:
            self.query = self.all_results.loc['query', 'Metadata']
        except:
            self.query = None
        
        # If WhoIs search has been successful, assigns entities result as an attribute
        try:
            self.entities = self.all_results.loc['entities', 'Metadata']
        except:
            self.entities = None
        
        # If WhoIs search has been successful, assigns raw result as an attribute
        try:
            self.raw = self.all_results.loc['raw', 'Metadata']
        except:
            self.raw = None
        
        # If WhoIs search has been successful, assigns network result as an attribute
        try:
            self.network = pd.DataFrame(pd.Series(self.all_results.loc['network', 'Metadata']), columns = ['Metadata'])
            self.network.index.name = 'Category'
        except:
            self.network = None
        
        # If WhoIs search has been successful, parses network events result and assigns as an attribute
        try:
            
            self.network_events = pd.DataFrame(columns = ['action', 'timestamp', 'actor'])
            events = self.network.loc['events', 'Metadata']
            
            for i in events:
                df = pd.DataFrame(pd.Series(i)).T
                self.network_events = pd.concat([self.network_events, df])
            
            self.network_events = self.network_events.reset_index().drop('index', axis=1)
            self.all_results.at['network_events', 'Metadata'] = events
            
        except:
            self.network_events = None
        
        # If WhoIs search has been successful, parses network notices result and assigns as an attribute
        try:
            
            self.network_notices = pd.DataFrame(columns = ['title', 'description', 'links'])
            notices = self.network.loc['notices', 'Metadata']
            
            for i in notices:
                df = pd.DataFrame(pd.Series(i)).T
                self.network_notices = pd.concat([self.network_notices, df])
            
            self.network_notices = self.network_notices.reset_index().drop('index', axis=1)
            self.all_results.at['network_notices', 'Metadata'] = notices

        except:
            self.network_notices = None
        
        # If WhoIs search has been successful, parses objects result and assigns as an attribute
        try:
            objects_series = pd.Series(self.all_results.loc['objects', 'Metadata'])
            self.objects = pd.DataFrame(dtype = object)
            
            for i in objects_series:
                obj_df = pd.DataFrame(pd.Series(i)).T
                self.objects = pd.concat([self.objects, obj_df]).reset_index().drop('index', axis=1)
            
        except:
            self.objects = None
        
        # If WhoIs search has been successful, parses contacts result and assigns as an attribute
        try:
            contacts_series = self.objects['contact']
    
            self.contacts = pd.DataFrame(dtype = object)
            
            for i in contacts_series:
                contact_df = pd.DataFrame(pd.Series(i)).T
                try:
                    contact_df.loc[0, 'address'] = contact_df.loc[0, 'address'][0]['value']
                except:
                    None
                
                try:
                    contact_df.loc[0, 'phone'] = contact_df.loc[0, 'phone'][0]['value']
                except:
                    None
                
                try:
                    contact_df.loc[0, 'email'] = contact_df.loc[0, 'email'][0]['value']
                except:
                    None
                
                self.contacts = pd.concat([self.contacts, contact_df]).reset_index().drop('index', axis=1)
                self.all_results.at['contacts', 'Metadata'] = contacts_series.to_list()
        except:
            self.contacts = None
        
        # Cleaning up results dataframe
        self.all_results = self.all_results.replace(np.nan, None)
        self.all_results = self.all_results.reset_index().drop_duplicates(subset = 'Category').set_index('Category')
                        
            
    def domain_whois(self, domain: str, ip_address: str):
    
        """
        Performs a WhoIs lookup on a domain.
        """

        # Checking if domain address given is valid; trying to retrieve domain if not
        if domain != None:
                if is_domain(domain) == False:
                    try:
                        domain = domain_splitter(domain)
                    except:
                        pass

                    if is_domain(domain) == False:
                        domain = None

        # Checking if IP address given is valid
        if ip_address != None:
            if is_ip_address(ip_address) == False:
                ip_address = None

        # If only an IP address has been given, tries to retrieve associated domain address
        if (domain == None) and (ip_address != None):
            try:
                domain = domain_from_ip(ip_address)['domain_name']
            except:
                domain = None
                return

            # Using recursion, runs a WhoIs lookup on the domain that is found
            domain_whois(self, domain = domain, ip_address = ip_address)

        # Running WhoIs lookup on the domain given
        if domain != None:

                # Trying to run WhoIs search on domain
                try:
                    # Creating dataframe for output and inputting WhoIs result
                    df = pd.DataFrame(pd.Series(whois.whois(domain)), columns = ['Metadata'])
                    df.index.name = 'Category'
                    self.all_results = pd.concat([self.all_results, df])
                    self.all_results.loc['domain', 'Metadata'] = domain

                except:
                    pass

                # Trying to retrieve associated IP address if none given
                try:
                    if ip_address == None:
                        try:
                            ip_address = ip_from_domain(domain)
                        except:
                            ip_address = None

                        self.ip_address = ip_address
                        self.all_results.loc['ip_address', 'Metadata'] = ip_address

                    else:
                        pass

                except:
                    pass
    
    
    def ip_whois(self, domain: str, ip_address: str):

        """
        Performs a WhoIs lookup on an IP address.
        """

        # Checking if domain address given is valid; trying to retrieve domain if not
        if domain != None:
                if is_domain(domain) == False:
                    try:
                        domain = domain_splitter(domain)
                    except:
                        pass

                    if is_domain(domain) == False:
                        domain = None

        # Checking if IP address given is valid
        if ip_address != None:

            ip_address = ip_address.strip().strip('/').strip()
            if is_ip_address(ip_address) == False:
                ip_address = None

        # If only a domain has been given, tries to retrieve associated IP address
        if (domain != None) and (ip_address == None):
            try:
                ip_address = ip_from_domain(domain)
                self.ip_address = ip_address

                # Using recursion, runs a WhoIs lookup on the IP address that is found
                ip_whois(self, domain = domain, ip_address = ip_address)
            except:
                pass

        # Running WhoIs lookup on the IP address given
        if ip_address != None:
                
                # Retrieving domain if none already given/found
                if domain == None:
                    try:
                        domain = domain_from_ip(ip_address)['domain_name']
                    except:
                        domain = None
                    self.domain = domain
                
                # Running domain WhoIs on domain
                try:
                    domain_whois(self, domain = domain, ip_address = ip_address)
                except:
                    pass
                
                # Formatting results as pandas.DataFrame
                try:
                    domain_reverse_search = pd.DataFrame(domain_from_ip(ip_address), columns = ['Metadata'])
                    domain_reverse_search.index.name = 'Category'
                    self.all_results = pd.concat([self.all_results, domain_reverse_search])
                except:
                    pass
                
                # Running WhoIs lookup on IP address
                try:
                    self.IPWhois_obj = IPWhois(ip_address)
                except:
                    pass
                
                # Retrieving geocode associated with IP address using Geopy
                try:
                    self.ip_geocode = get_ip_geocode(ip_address)
                    self.all_results.at['ip_geocode', 'Metadata'] = self.ip_geocode
                except:
                    pass
                
                # Retrieving coordinates associated with IP address using Geopy
                try:
                    self.ip_coordinates = get_ip_coordinates(ip_address)
                    self.all_results.at['ip_coordinates', 'Metadata'] = self.ip_coordinates
                except:
                    pass
                
                # Retrieving location associated with IP address using Geopy
                try:
                    self.ip_location = get_ip_physical_location(ip_address)
                    self.all_results.at['ip_location', 'Metadata'] = self.ip_location
                except:
                    pass
                
                # Assigning results to WhoisResult object
                try:
                    self.all_results.at['ip_address', 'Metadata'] = ip_address
                except:
                    None

        else:
            self.IPWhois_obj = None



    def __repr__(self):
        
            """Controls how WhoIsResult objects are represented in string form."""
        
            return str(self.all_results)

    def contents(self):
            
            """Returns a list of object contents."""
            
            return list(self.__dict__.keys())

    def copy(self):
            
            """Creates a copy of object."""
            
            return copy.deepcopy(self)
    
def domain_whois(domain: str = 'request_input'):
    
    """
    Performs a WhoIs lookup on a domain address.
    """
    
    # Requesting domain address from user input if none given
    if domain == 'request_input':
        domain = input('Domain: ')
    
    # Raising error if domain given is not a string
    if type(domain) != str:
        raise TypeError('Domain must be a string')
    
    # Checking if domain is valid; if not, tries to extract domain
    if is_domain(domain) == False:
        domain = domain_splitter(domain)
    
     # Re-checking if domain is valid; if true, creates WhoisResult object
    if is_domain(domain) == True:
        return WhoisResult(domain = domain)
    
    else:
        return None

def domains_whois(domains_list: List[str]):
    
    """
    Performs a WhoIs lookup on a list or set of domain addresses.
    """
    
    # Creating output dataframe
    output_df = pd.DataFrame(dtype=object)
    
    # Iterating through domains
    for domain in domains_list:
        
        if domain == None:
            df = pd.DataFrame()
            
        else:
            # If domain is valid, running WhoIs lookup using WhoisResult class
            if is_domain(domain) == True:
                try:
                    df = domain_whois(domain = domain).all_results.T
                except:
                    continue
            else:
                # If domain is not valid, trying to retrieve domain
                try:
                    domain = domain_splitter(domain)
                except:
                    continue
                
                # If domain is valid, running WhoIs lookup using WhoisResult class
                if is_domain(domain) == True:
                    try:
                        df = domain_whois(domain = domain).all_results.T
                    except:
                        continue
        
        # Concatanating results dataframes
        output_df = pd.concat([output_df, df])
    
    # Cleaning and reformatting output dataframe
    output_df = output_df.replace(np.nan, None).reset_index().drop('index', axis=1)
    output_df = output_df.set_index('domain')
    output_df.index.rename('domain')
    output_df.columns.name = None
    
    return output_df
        

def ip_whois(ip_address: str = 'request_input'):
    
    """
    Performs a WhoIs lookup on an IP address.
    """
    
    # Requesting domain address from user input if none given
    if ip_address == 'request_input':
        ip_address = input('IP address: ')
    
    # Cleaning IP address
    ip_address = ip_address.strip().strip('/').strip()
    
    # Checking if IP address is valid; if true, running WhoIs lookup using WhoisResult class
    if is_ip_address(ip_address) == True:
        return WhoisResult(ip_address = ip_address)
    
    else:
        return None

def ips_whois(ip_addresses = List[str]):
    
    """
    Performs a WhoIs lookup on a list of IP addresses.
    """
    
    # Creating output dataframe
    output_df = pd.DataFrame(dtype=object)
    
    # Iterating through IP addresses
    for ip in ip_addresses:
        
        if ip == None:
            df = pd.DataFrame()
        
        # If IP address is inputted, tries to run a WhoIs lookup using WhoisResult class
        else:
        
            try:
                df = ip_whois(ip_address = ip).all_results.T
            except:
                df = pd.DataFrame()
        
        # Concatanating results dataframes
        output_df = pd.concat([df, output_df])
    
    # Cleaning and reformatting output dataframe
    output_df = output_df.replace(np.nan, None).reset_index().drop('index', axis=1)
    output_df = output_df.set_index('ip_address')
    output_df.index.rename('ip_address')
    output_df.columns.name = None
    
    return output_df


def lookup_whois(string: str = 'request_input'):
    
    """
    Performs a WhoIs lookup on an inputted domain or IP address.
    """
    
    # Requesting string from user input if none given
    if string == 'request_input':
        string = input('Search query: ')
    
    # Raising error if input is not a string object
    if type(string) != str:
        raise TypeError('Search query must be a string')
    
    # Making a copy of string
    original_string = copy.deepcopy(string)
    
    # Cleaning string
    string = string.strip().strip('/').strip('\\').strip('#').strip('.').strip('-').strip()
    
    # If string is an IP address, runs IP WhoIs lookup
    if is_ip_address(string) == True:
        try:
            return ip_whois(string)
        except:
            return
    
    # If string is not an IP address and is not a valid domain, tries to retrieve domain
    if is_domain(string) == False:
        string = domain_splitter(string)
    
    # If string is a domain, runs domain WhoIs lookup
    if is_domain(string) == True:
        try:
            return domain_whois(domain = string)
        except:
            return
    
    else:
        raise ValueError(f'Search query "{original_string}" contains neither a valid domain name nor a valid IP address')
    
    
    

## Functions for opening links and addresses

def open_url(url = 'request_input'):
    
    """
    Opens URL in the default web browser.
    """
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Opening browser
    return webbrowser.open(url)

def open_doi(doi = 'request_input'):

    if doi == 'request_input':
        doi = input('DOI: ')
    
    url = 'https://doi.org/' + doi
    
    return webbrowser.open(url)

def open_urls_list(urls: list):
    
    """
    Opens list of URLs in the default web browser.
    """
    
    # Iterating through URLs list
    for link in urls:
        open_url(link)

def open_url_source(url = 'request_input'):
    
    """
    Opens URL's source code in the default web browser.
    """
    
    # Requesting URL from user input if none given
    if url == 'request_input':
        url = input('URL: ')
    
    # Editing URL to view source
    url = 'view-source:' + url
    
    # Opening browser
    open_url(url)

def regex_check_then_open_url(url: str):
    
        """
        Checks if string is a web address or file path. If true, opens using default web browser.
        """
    
    # Regex objects to check type

        www = re.compile(r'www\.', re.IGNORECASE)
        https = re.compile(r'https://', re.IGNORECASE)
        http = re.compile(r'http://', re.IGNORECASE)
        file_address = re.compile(r'file://', re.IGNORECASE)
        win_drive = re.compile(r'[A-Za-z]:\\', re.IGNORECASE)
        unc = re.compile(r'\\\\', re.IGNORECASE)
        rel = re.compile(r'\.\\', re.IGNORECASE)
        mac = re.compile(r'~/', re.IGNORECASE)
        mac_user = re.compile(r'/Users', re.IGNORECASE)
        unix_par = re.compile(r'\.\.', re.IGNORECASE)
        unix_home = re.compile(r'~', re.IGNORECASE)

         # Checking for matches
        www_match = www.search(url)
        https_match = https.search(url)
        http_match = http.search(url)
        file_address_match = file_address.search(url)
        win_drive_match = win_drive.search(url)
        unc_match = unc.search(url)
        rel_match = rel.search(url)
        mac_match = mac.search(url)
        mac_user_match = mac_user.search(url)
        unix_p_match = unix_par.search(url)
        unix_h_match = unix_home.search(url)
        
        # Checking if string is a file path; if true, opens in web browser
        if (
                (file_address_match != None) == True
                or ((win_drive_match != None) == True)
                or ((unc_match != None) == True)
                or ((rel_match != None) == True)
                or ((mac_match != None) == True)
                or ((mac_user_match != None) == True)
                or ((unix_p_match != None) == True)
                or ((unix_h_match != None) == True)
                ):
                is_filepath = 'True'
                url = 'file://' + url
                return webbrowser.open(url, new=1)
        
        # Checking if string is a web address; if true, opens in web browser
        if (
                ((www_match != None) == True)
                and (https_match == None)
                and (http_match == None)
                ):
                is_webaddress = 'True'
                url = 'https://' + url
                return webbrowser.open(url)

        # Checking if string is a web address; if true, opens in web browser
        if (((https_match != None) == True)
                or ((http_match != None) == True)
                ):
                is_webaddress = 'True'
                return webbrowser.open(url)
        
        # If checks have failed, searches for string on Google
        if ((
            www_match == None
            and https_match == None
            and http_match == None
            and file_address_match == None
            and win_drive_match == None
            and unc_match == None
            and rel_match == None
            and mac_match == None
            and mac_user_match == None
            and unix_p_match == None
            and unix_h_match == None)
            and (',' or ', ') not in url
            ):
                query = quote(url)
                url_base = 'https://www.google.com/search?hl=en&q='
                url = url_base + query
                return webbrowser.open(url)

def url_to_valid_attr_name(url: str) -> str:
    
    """
    Converts URL to a string which can be used as a Python object attribute name.
    """
    
    name = url.replace('https://', '').replace('http://', '').replace('www.', '')
    name = name.replace('.', '_').replace('~', '_').replace('/', '_').replace('&', '_').replace('#', '_').replace('@', '_at_')
    name = name.replace('(', '').replace(')', '').replace('{', '').replace('}', '').replace('%', '').replace('!', '').replace('?', '').replace('*', '').replace(':', '')
    name = name.replace('<', '').replace('>', '').replace('-', '').replace('+', '_').replace('=', '_')
    name = name.replace('"', '').replace("'", "")
    name = name.strip().lower()
    
    return name


from ..utils.cleaners import deduplicate
from ..datasets.stopwords.stopwords import all_stopwords
from ..importers.crossref import lookup_funder

from .entities import Entity, Entities


import pandas as pd
import numpy as np

from geopy.geocoders import Nominatim # type: ignore
from nltk.tokenize import word_tokenize # type: ignore

def generate_affiliation_id(affiliation_data: pd.Series):

        affiliation_data = affiliation_data.copy(deep=True).dropna().astype(str).str.lower()

        affiliation_id = 'AFFIL:'

        if 'name' in affiliation_data.index:
            name = affiliation_data['name']
        else:
            name = ''

        if (name == None) or (name == ''):
            name = 'no_name_given'
        
        if name != 'no_name_given':
            name = name.strip().lower()
            name_tokens = list(word_tokenize(name))
            name_tokens = [i for i in name_tokens if i not in all_stopwords]
            name_first_3 = name_tokens[:3]
            name_last = name_tokens[-1]

            if name_last in name_first_3:
                name_last = ''
            
            name_shortened = '-'.join(name_first_3) + '-' + name_last

        else:
            name_shortened = name

        affiliation_id = affiliation_id + '-' + name_shortened.lower()

        if 'location' in affiliation_data.index:
            location = affiliation_data['location']
            if type(location) == str:
                location = location.split(',')
                location_shortened = location[0].strip()
                if location[0] != location[-1]:
                    location_shortened = location_shortened + '-' + location[-1].strip()
            else:
                location_shortened = ''
        else:
            location_shortened = ''
        
        affiliation_id = affiliation_id + '-' + location_shortened.lower()

        uid = ''

        if 'uri' in affiliation_data.index:
            uid = affiliation_data['uri']

            if (uid == None) or (uid == 'None') or (uid == ''):
                if 'crossref_id' in affiliation_data.index:
                    uid = affiliation_data['crossref_id']

                    if (uid == None) or (uid == 'None') or (uid == ''):
                        if 'website' in affiliation_data.index:
                            uid = affiliation_data['website']

                            if (uid == None) or (uid == 'None') or (uid == ''):
                                uid = ''
        
        uid_shortened = uid.replace('https://', '').replace('http://', '').replace('www.', '').replace('dx.','').replace('doi.org/','').replace('user=','')[:24]

        affiliation_id = affiliation_id + '-' + uid_shortened
        affiliation_id = affiliation_id.replace('AFFIL:-', 'AFFIL:').replace("'s", '').replace('\r', '').replace('\n', '').replace('.', '').replace("'", "").replace('"', '').replace('(','').replace(')','').replace('`','').replace('’','').replace('--', '-').replace('AFFIL:-', 'AFFIL:').replace(' ', '-').strip('-')

        return affiliation_id

class Affiliation(Entity):

    """
    This is a Affiliation object. It is designed to store data about an organisation that an author is affiliated with.
    
    Parameters
    ----------
    
    
    Attributes
    ----------
    """

    def __init__(self,
                 affiliation_id: str = None, # type: ignore
                 name: str = None, # type: ignore
                 location: str = None, # type: ignore
                 address: str = None, # type: ignore
                 email: str = None, # type: ignore
                 uri: str = None, # type: ignore
                 crossref_id: int = None, # type: ignore
                 website: str = None,  # type: ignore
                 other_links = [], # type: ignore
                 use_api: bool = False
                 ):
        
        """
        Initialises affiliation instance.
        
        Parameters
        ----------
        """

        super().__init__()

        orig_name_data = name

        if type(name) == str:
            name = name.strip().replace('&amp;', '&').replace('\n', '').replace('\n', ' ')
        
            if (location is None) and (address is None):

                name_split = name.split(',')
                name_split = [i.strip() for i in name_split]

                address = name
                location = name_split[-1]

                if (len(name_split) > 1) and (('department' in name_split[0].lower()) or ('institute of' in name_split[0].lower())):
                    name = name_split[0] + ', ' + name_split[1]
                else:
                    name = name_split[0]
        
        if type(name) == list:

            if (location is None) and (address is None):
                address = ', '.join(name)
                location = name[-1]
            
            name = name[0]

        if type(other_links) == str:
            other_links = other_links.replace('[','').replace(']','').split(',')
            other_links = [i.strip() for i in other_links]
        

        self.summary = pd.DataFrame(columns = [
                                'affiliation_id',
                                'name',
                                'location',
                                'address',
                                'email',
                                'uri',
                                'crossref_id',
                                'website',
                                'other_links'
                                ],
                                dtype = object)
        
        
        self.summary.loc[0] = pd.Series(dtype=object)
        self.summary.loc[0, 'affiliation_id'] = affiliation_id
        self.summary.loc[0, 'name'] = name
        self.summary.loc[0, 'location'] = location
        self.summary.loc[0, 'address'] = address
        self.summary.loc[0, 'email'] = email
        self.summary.loc[0, 'uri'] = uri
        self.summary.loc[0, 'crossref_id'] = crossref_id
        self.summary.loc[0, 'website'] = website
        self.summary.at[0, 'other_links'] = other_links

        if use_api == True:
            
            if (type(orig_name_data) != str) and (type(orig_name_data) != list):
                orig_name_data = str(orig_name_data).strip().replace('{','').replace('}','').replace('[','').replace(']','')

            if type(orig_name_data) == str:
                orig_name_data = orig_name_data.split(',')
                orig_name_data = [i.strip() for i in orig_name_data]

            if type(orig_name_data) == list:
                location_data = orig_name_data[-3:]
                location_data = ', '.join(location_data)
            else:
                location_data = ''

            geolocator = Nominatim(user_agent="location_app")

            try:
                loc = geolocator.geocode(orig_name_data)
            except:
                loc = None
            
            if loc != None:
                self.summary.loc[0, 'address'] = loc.address

                if self.summary.loc[0, 'name'] == None:
                    self.summary.loc[0, 'name'] = loc.name
                
                if self.summary.loc[0, 'location'] == None:
                    self.summary.loc[0, 'location'] = loc.display_name

        self.update_id()

    def generate_id(self):

        affiliation_data = self.summary.loc[0]

        affiliation_id = generate_affiliation_id(affiliation_data) # type: ignore
        return affiliation_id

    def update_id(self):

        current_id = str(self.summary.loc[0, 'affiliation_id'])

        if (current_id == None) or (current_id == 'None') or (current_id == '') or (current_id == 'AFFIL:000') or ('no_name_given' in current_id):
            auth_id = self.generate_id()
            self.summary.loc[0, 'affiliation_id'] = auth_id
        
    
    def __getitem__(self, key):
        
        """
        Retrieves affiliation attribute using a key.
        """
        
        if key in self.__dict__.keys():
            return self.__dict__[key]

        if key in self.summary.columns:
            return self.summary.loc[0, key]

    def __repr__(self) -> str:
        
        return str(self.summary.loc[0, 'name'])

    def has_uri(self) -> bool:

        uri = self.summary.loc[0, 'uri']

        if (type(uri) == str) and (uri != ''):
            return True
        else:
            return False

    def add_dict(self, data: dict):

        if 'name' in data.keys():
            name = data['name']
            self.summary.loc[0, 'name'] = name
        
        if 'location' in data.keys():
            location = data['location']
            self.summary.loc[0, 'location'] = location

        if 'address' in data.keys():
            address = data['address']
            self.summary.loc[0, 'address'] = address
        
        if 'crossref_id' in data.keys():
            crossref_id = data['crossref_id']
            self.summary.loc[0, 'crossref_id'] = crossref_id

        if 'DOI' in data.keys():
            uri = data['DOI'].replace('http', '').replace('https', '').replace('dx.', '').replace('doi.org/', '').strip()
            self.summary.loc[0, 'uri'] = 'https://doi.org/' + uri
        else:
            if 'uri' in data.keys():
                uri = data['DOI'].replace('http', '').replace('https', '').replace('dx.', '').replace('doi.org/', '').strip()
                self.summary.loc[0, 'uri'] = 'https://doi.org/' + uri

        if 'url' in data.keys():
            website = data['url']
            self.summary.loc[0, 'website'] = website
        else:
            if 'link' in data.keys():
                website = data['link']
                self.summary.loc[0, 'website'] = website
            else:
                if 'website' in data.keys():
                    website = data['website']
                    self.summary.loc[0, 'website'] = website

    def from_dict(data: dict, use_api=False): # type: ignore

        if 'name' in data.keys():
            name = data['name']
        else:
            name = None
        
        if 'location' in data.keys():
            location = data['location']
        else:
            location = None

        if 'address' in data.keys():
            address = data['address']
        else:
            address = None
        
        if 'email' in data.keys():
            email = data['email']
        else:
            email = None

        if 'crossref_id' in data.keys():
            crossref_id = data['crossref_id']
        else:
            crossref_id = None

        if 'DOI' in data.keys():
            uri = data['DOI'].replace('http', '').replace('https', '').replace('dx.', '').replace('doi.org/', '').strip()
            uri = 'https://doi.org/' + uri
        else:
            if 'doi' in data.keys():
                uri = data['doi'].replace('http', '').replace('https', '').replace('dx.', '').replace('doi.org/', '').strip()
                uri = 'https://doi.org/' + uri
            else:
                if 'URI' in data.keys():
                    uri = data['URI'].replace('http', '').replace('https', '').replace('dx.', '').replace('doi.org/', '').strip()
                    uri = 'https://doi.org/' + uri
                else:
                    if 'uri' in data.keys():
                        uri = data['DOI'].replace('http', '').replace('https', '').replace('dx.', '').replace('doi.org/', '').strip()
                        uri = 'https://doi.org/' + uri
                    else:
                        uri = None

        if 'URL' in data.keys():
            website = data['URL']
        else:
            if 'url' in data.keys():
                website = data['url']
            else:
                if 'link' in data.keys():
                    website = data['link']
                else:
                    if 'website' in data.keys():
                        website = data['website']
                    else:
                        website = None

        affiliation = Affiliation(name=name, location=location, address=address, email=email, uri=uri, crossref_id=crossref_id, website=website, use_api=use_api) # type: ignore

        return affiliation
        
    def add_series(self, series: pd.Series):
        self.summary.loc[0] = series

    def from_series(data: pd.Series): # type: ignore
        affiliation = Affiliation()
        affiliation.add_series(data)

        return affiliation

    def add_dataframe(self, dataframe: pd.DataFrame):
        series = dataframe.loc[0]
        self.add_series(series) # type: ignore

    def from_dataframe(data: pd.DataFrame): # type: ignore
        affiliation = Affiliation()
        affiliation.add_dataframe(data)

        return affiliation

    def import_crossref_result(self, crossref_result: pd.Series):
        
        if 'name' in crossref_result.index:
            name = crossref_result['name']
        else:
            name = self.summary.loc[0, 'name']

        if 'location' in crossref_result.index:
            location = crossref_result['location']
        else:
            location = self.summary.loc[0, 'location']

        if 'email' in crossref_result.index:
            email = crossref_result['email']
        else:
            email = self.summary.loc[0, 'email']

        if 'uri' in crossref_result.index:
            uri  =crossref_result['uri']
        else:
            uri = self.summary.loc[0, 'uri']

        if 'id' in crossref_result.index:
            crossref_id = crossref_result['id']
        else:
            crossref_id = self.summary.loc[0, 'crossref_id']
        
        self.summary.loc[0, 'name'] = name
        self.summary.loc[0, 'location'] = location
        self.summary.loc[0, 'email'] = email
        self.summary.loc[0, 'uri'] = uri
        self.summary.loc[0, 'crossref_id'] = crossref_id

    def from_crossref_result(crossref_result: pd.Series, use_api: bool = False): # type: ignore
        
        if 'name' in crossref_result.index:
            name = crossref_result['name']
        else:
            name = None

        if 'location' in crossref_result.index:
            location = crossref_result['location']
        else:
            location = None

        if 'email' in crossref_result.index:
            email = crossref_result['email']
        else:
            email = None

        if 'uri' in crossref_result.index:
            uri  =crossref_result['uri']
        else:
            uri = None

        if 'id' in crossref_result.index:
            crossref_id = crossref_result['id']
        else:
            crossref_id = None

        affiliation = Affiliation(name=name, location=location, email=email, uri=uri, crossref_id=crossref_id, use_api=use_api) # type: ignore

        return affiliation

    def import_crossref(self, crossref_id: str, timeout = 60):

        res = lookup_funder(crossref_id, timeout)
        self.import_crossref_result(res.loc[0]) # type: ignore

    def from_crossref(crossref_id: str, use_api=True, timeout = 60): # type: ignore
        res = lookup_funder(crossref_id, timeout)
        affiliation = Affiliation.from_crossref_result(crossref_result=res, use_api=use_api) # type: ignore

        return affiliation
    
    def import_uri(self, uri: int, timeout = 60):
        uri_str = str(uri)
        res = lookup_funder(uri_str, timeout)
        self.import_crossref_result(res.loc[0]) # type: ignore

    def from_uri(uri: int, use_api=True, timeout = 60): # type: ignore
        uri_str = str(uri)
        res = lookup_funder(uri_str, timeout)
        affiliation = Affiliation.from_crossref_result(crossref_result=res, use_api=use_api) # type: ignore

        return affiliation
    
    def update_address(self):
        
        if self.summary.loc[0, 'name'] != None:
            name = str(self.summary.loc[0, 'name']).strip().replace('{','').replace('}','').replace('[','').replace(']','').replace(',',' ').replace('  ',' ').strip()
        else:
            name = ''
        
        if self.summary.loc[0, 'location'] != None:
            location = str(self.summary.loc[0, 'location']).strip().replace('{','').replace('}','').replace('[','').replace(']','').replace('  ',' ').strip()
        else:
            location = ''

        if self.summary.loc[0, 'address'] != None:
            address = str(self.summary.loc[0, 'address']).strip().replace('{','').replace('}','').replace('[','').replace(']','').replace('  ',' ').strip()
        else:
            address = ''

        if name not in address:
            address = name + ', ' + address
        
        if (address == '') or (address == ', '):
            address = location

        geolocator = Nominatim(user_agent="location_app")

        try:
                loc = geolocator.geocode(address)
        except:
                loc = None
            
        if loc != None:
                
                self.summary.loc[0, 'address'] = loc.address

                if self.summary.loc[0, 'name'] == None:
                    self.summary.loc[0, 'name'] = loc.name
                
                if self.summary.loc[0, 'location'] == None:
                    self.summary.loc[0, 'location'] = loc.display_name


    def update_from_crossref(self, timeout = 60):

        uid = self.summary.loc[0,'crossref_id']
        if uid == None:
            uid = self.summary.loc[0,'uri']
            if uid == None:
                uid = ''

        res = lookup_funder(funder_id = uid, timeout = timeout) # type: ignore
        if len(res) > 0:
            self.import_crossref_result(res.loc[0]) # type: ignore

    def update_from_uri(self, timeout = 60):

        uid = self.summary.loc[0, 'uri']
        if uid == None:
            uid = self.summary.loc[0, 'crossref']
            if uid == None:
                uid = ''

        res = lookup_funder(funder_id = uid, timeout = timeout) # type: ignore
        if len(res) > 0:
            self.import_crossref_result(res.loc[0]) # type: ignore
   
class Affiliations(Entities):

    """
    This is a Affiliations object. It contains a collection of Affiliations objects and compiles data about them.
    
    Parameters
    ----------
    
    
    Attributes
    ----------
    """

    def __init__(self, affiliations_data = None):
        
        """
        Initialises Affiliations instance.
        
        Parameters
        ----------
        """

        super().__init__()

        self.summary = pd.DataFrame(columns = 
                                [
                                'affiliation_id',
                                'name',
                                'location',
                                'address',
                                'email',
                                'uri',
                                'crossref_id',
                                'website',
                                'other_links'
                                ],
                                dtype = object)
        

        self.all = dict()

        self.data = []

        self.data.append(affiliations_data)

        if (type(affiliations_data) == list) and (type(affiliations_data[0]) == Affiliation):

            for a in affiliations_data:
                affil_details = a.summary.copy(deep=True)
                affil_id = affil_details.loc[0, 'affiliation_id']
                self.summary = pd.concat([self.summary, affil_details])
                self.all[affil_id] = a

            self.summary = self.summary.reset_index().drop('index',axis=1)

        else:

            if (type(affiliations_data) == list) and (type(affiliations_data[0]) == dict):

                for i in affiliations_data:
                    a = Affiliation.from_dict(i) # type: ignore
                    affil_id = a.summary.loc[0, 'affiliation_id']
                    affil_details = a.summary.copy(deep=True)
                    self.summary = pd.concat([self.summary, affil_details])
                    self.all[affil_id] = a

                self.summary = self.summary.reset_index().drop('index',axis=1)

            else:

                if type(affiliations_data) == dict:
                    
                    values = list(affiliations_data.values())

                    if type(values[0]) == Affiliation:

                        for a in affiliations_data.keys():
                            affil_id = a.summary.loc[0, 'affiliation_id']
                            affil_details = a.summary.copy(deep=True)
                            self.summary = pd.concat([self.summary, affil_details])
                            self.all[affil_id] = a

                        self.summary = self.summary.reset_index().drop('index',axis=1)
                
        self.update_ids()

    def __getitem__(self, key):
        
        """
        Retrieves affiliations attribute using a key.
        """
        
        if key in self.__dict__.keys():
            return self.__dict__[key]
        
        if key in self.all.keys():
            return self.all[key]

        if key in self.summary.columns:
            return self.summary[key]
        
        if (type(key) == int) and (key <= len(self.data)):
            return self.data[key]
    
    def __repr__(self) -> str:

        alphabetical = str(self.summary['name'].sort_values().to_list()).replace('[','').replace(']','')
        return alphabetical
    
    def __len__(self) -> int:
        return len(self.all.keys())

    def drop_empty_rows(self):

        ignore_cols = ['affiliation_id', 'address', 'email', 'other_links']

        df = self.summary.copy(deep=True)
        df['name'] = df['name'].replace('no_name_given', None)
        df = df.dropna(axis=0, how='all')
        drop_cols = [c for c in df.columns if c not in ignore_cols]
        df = df.dropna(axis=0, how='all', subset=drop_cols).reset_index().drop('index', axis=1)

        self.summary = df

        return self
    
    def remove_duplicates(self, drop_empty_rows = True, sync = False):

        if drop_empty_rows == True:
            self.drop_empty_rows()
        
        df = self.summary.copy(deep=True)
        df['uri'] = df['uri'].str.replace('http://', '', regex=False).str.replace('https://', '', regex=False).str.replace('wwww.', '', regex=False).str.replace('dx.', '', regex=False).str.replace('doi.org/', '', regex=False).str.strip('/')
        
        df = df.sort_values(by = ['uri', 'crossref_id', 'website', 'name', 'location', 'address']).reset_index().drop('index', axis=1)
        self.summary = deduplicate(self.summary)

        if sync == True:
            self.sync_details(drop_duplicates=False, drop_empty_rows=False)

        return self


    def sync_all(self, drop_duplicates = False, drop_empty_rows=False):

        for i in self.all.keys():
            affil = self.all[i]
            affil.update_id()
            series = affil.summary.copy(deep=True).loc[0]
            all = self.summary.copy(deep=True).astype(str)
            indexes = all[all['affiliation_id'] == i].index.to_list()
            if len(indexes) > 0:
                affil_index = indexes[0]
                all_copy = self.summary.copy(deep=True)
                all_copy.loc[affil_index] = series
                self.summary = all_copy
        
        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

    def sync_details(self, drop_duplicates = False, drop_empty_rows=False):

        self.update_ids(sync=False)

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        for i in self.summary.index:

            a_data = self.summary.loc[i]
            a_id = a_data['affiliation_id']

            if a_id != None:
                a = Affiliation.from_series(a_data) # type: ignore
                self.all[a_id] = a

            else:
                a_id = generate_affiliation_id(a_data)
                a_data['affiliation_id'] = a_id
                a = Affiliation.from_series(a_data) # type: ignore
                self.all[a_id] = a
        
        keys = list(self.all.keys())
        for key in keys:
            a_ids = self.summary['affiliation_id'].to_list()
            if key not in a_ids:
                del self.all[key]

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)


    def sync(self, drop_duplicates = False, drop_empty_rows=False):
        
        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        all_len = len(self.summary)
        details_len = len(self.all)

        if all_len > details_len:
            self.sync_details(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
            return
        else:
            if details_len > all_len:
                self.sync_all(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
                return
            else:
                self.sync_details(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
                self.sync_all(drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)
                return


    def merge(self, affiliations, drop_duplicates = False, drop_empty_rows=True):

        left = self.summary.copy(deep=True)
        right = affiliations.summary.copy(deep=True)
        
        merged = pd.concat([left, right])

        self.summary = merged.drop_duplicates(subset=['affiliation_id', 'name', 'location'], ignore_index=True)

        for i in affiliations.all.keys():
            if i not in self.all.keys():
                self.all[i] = affiliations.all[i]

        left_data = self.data
        right_data = affiliations.data

        if left_data == None:
                left_data = []
            
        if right_data == None:
                right_data = []

        if (type(left_data) == Affiliation) or (type(left_data) == str):
                left_data = [left_data]
            
        if (type(right_data) == Affiliation) or (type(right_data) == str):
                right_data = [right_data]
            
        if type(left_data) == dict:
                left_data = list(left_data.values())
            
        if type(right_data) == dict:
                right_data = list(right_data.values())

        merged_data = left_data + right_data # type: ignore
        merged_data = pd.Series(merged_data).value_counts().index.to_list()

        self.data = merged_data

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        self.update_ids()

        return self


    def add_affiliation(self, affiliation: Affiliation =  None, name: str = None, uri: str = None, crossref_id: int = None, data = None, use_api = True, drop_duplicates = False, drop_empty_rows=False): # type: ignore

        if type(affiliation) == str:

            affiliation = Affiliation(name = affiliation, use_api = use_api)

            affiliation = None # type: ignore

        if data is not None:

            if type(data) == dict:
                affiliation = Affiliation.from_dict(data=data, use_api = use_api) # type: ignore
            
            else:
                if type(data) == pd.Series:
                    affiliation = Affiliation.from_series(data=data) # type: ignore
                
                else:
                    if type(data) == pd.DataFrame:
                        affiliation = Affiliation.from_dataframe(data=data) # type: ignore


        if affiliation is None:
            affiliation = Affiliation(name=name, uri=uri, crossref_id=crossref_id, use_api = use_api)

        if use_api == True:
            affiliation.update_from_crossref()
            affiliation.update_address()

        affiliation.update_id()

        affiliation_id = str(affiliation.summary.loc[0, 'affiliation_id'])

        # if affiliation_id in self.summary['affiliation_id'].to_list():
        #     all_copy = self.summary.copy(deep=True).astype(str)
        #     id_count = len(all_copy[all_copy['affiliation_id'].str.contains(affiliation_id)]) # type: ignore
        #     affiliation_id = affiliation_id + f'#{id_count + 1}'
        #     affiliation.summary.loc[0, 'affiliation_id'] = affiliation_id

        self.summary = pd.concat([self.summary, affiliation.summary])
        self.summary = self.summary.reset_index().drop('index', axis=1)

        self.all[affiliation_id] = affiliation

        if data is None:
            data = affiliation.summary.to_dict(orient='index')
        
        self.data.append(data)

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        self.update_ids()


    def add_list(self, affiliations_list: list, use_api: bool = False, drop_duplicates = False, drop_empty_rows=False):
        
        for i in affiliations_list:
            if type(i) == Affiliation:
                self.add_affiliation(affiliation = i, use_api=use_api)
            else:
                if type(i) == dict:
                    affil = Affiliation.from_dict(i)  # type: ignore
                    self.add_affiliation(affiliation = affil, use_api=use_api)
                else:
                    if type(i) == pd.Series:
                        affil = Affiliation.from_series(i)  # type: ignore
                        self.add_affiliation(affiliation = affil, use_api=use_api)
        
        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        self.update_ids()

    def update_ids(self, sync=False, drop_duplicates = False, drop_empty_rows=False):

        if sync == True:
            self.sync()

        for i in self.summary.index:
            
            data = self.summary.loc[i].copy(deep=True)
            old_id = self.summary.loc[i, 'affiliation_id']
            new_id = generate_affiliation_id(data)
            new_id = str(new_id)

            # if new_id in self.summary['affiliation_id'].to_list():
            #     df_copy = self.summary.copy(deep=True)
            #     df_copy = df_copy.astype(str)
            #     id_count = len(df_copy[df_copy['affiliation_id'].str.contains(new_id)])
            #     new_id = new_id + f'#{id_count + 1}'

            self.summary.at[i, 'affiliation_id'] = new_id
            if old_id in self.all.keys():
                self.all[new_id] = self.all[old_id]
                self.all[new_id].summary.loc[0, 'affiliation_id'] = new_id
                del self.all[old_id]

            else:
                affiliation = Affiliation.from_series(data) # type: ignore
                affiliation.summary.loc[0, 'affiliation_id'] = new_id
                self.all[new_id] = affiliation

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

    def update_addresses(self, sync=True, drop_duplicates = False, drop_empty_rows=False):

        if sync == True:
            self.sync(drop_duplicates=drop_duplicates,drop_empty_rows=drop_empty_rows)

        affiliation_ids = self.all.keys()

        for a in affiliation_ids:

            self.all[a].update_address()
            details = self.all[a].summary.loc[0]
            
            df_index = self.summary[self.summary['affiliation_id'] == a].index.to_list()[0]
            self.summary.loc[df_index] = details

            new_id = details['affiliation_id']
            if new_id != a:
                self.all[new_id] = self.all[a]
                del self.all[a]

        self.update_ids()

    def update_from_crossref(self, drop_duplicates = False, drop_empty_rows=False):

        affiliation_ids = self.all.keys()

        for a in affiliation_ids:

            self.all[a].update_from_crossref()
            details = self.all[a].summary.loc[0]
            
            df_index = self.summary[self.summary['affiliation_id'] == a].index.to_list()[0]
            self.summary.loc[df_index] = details

            new_id = details['affiliation_id']
            if new_id != a:
                self.all[new_id] = self.all[a]
                del self.all[a]

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        self.update_ids()


    def import_crossref_ids(self, crossref_ids: list, drop_duplicates = False, drop_empty_rows=False):

        for i in crossref_ids:

            auth = Affiliation.from_crossref(i) # type: ignore
            self.add_affiliation(affiliation = auth)

        if drop_empty_rows == True:
            self.drop_empty_rows()

        if drop_duplicates == True:
            self.remove_duplicates(drop_empty_rows=drop_empty_rows)

        self.update_ids()


    def from_crossref_ids(crossref_ids: list, drop_duplicates = False, drop_empty_rows=False): # type: ignore

        affiliations = Affiliations()
        affiliations.import_crossref_ids(crossref_ids, drop_empty_rows=drop_empty_rows, drop_duplicates=drop_duplicates)

        return affiliations

    def with_crossref(self):
        return self.summary[~self.summary['crossref_id'].isna()]
    
    def with_uri(self):
        return self.summary[~self.summary['uri'].isna()]

    def from_list(affiliations_list: list, use_api: bool = False, drop_duplicates = False, drop_empty_rows=False): # type: ignore
        affiliations = Affiliations()
        affiliations.add_list(affiliations_list, use_api=use_api, drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)

        return affiliations
    
    def import_crossref_result(self, crossref_result: pd.DataFrame, use_api = False, drop_duplicates = False, drop_empty_rows=False):

        for i in crossref_result.index:
            
            data = crossref_result.loc[i]
            affil = Affiliation.from_crossref_result(crossref_result=data, use_api=use_api) # type: ignore
            self.add_affiliation(affiliation = affil, use_api=use_api, drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)

        self.update_ids()


    def from_crossref_result(crossref_result: pd.DataFrame, use_api: bool = False, drop_duplicates = False, drop_empty_rows=False): # type: ignore

        affiliations = Affiliations()
        affiliations.import_crossref_result(crossref_result, use_api=use_api, drop_duplicates=drop_duplicates, drop_empty_rows=drop_empty_rows)

        return affiliations

def format_affiliations(affiliation_data, use_api = False, drop_duplicates = False, drop_empty_rows=True):
        
        result = Affiliations()

        affil_type = type(affiliation_data)

        if (
            (affil_type != pd.DataFrame) 
            and (affil_type != pd.Series) 
            and (affil_type != Affiliations) 
            and (affil_type != Affiliation) 
            and (affil_type != list) 
            and (affil_type != dict)
            ):
            result = Affiliations()
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result

        if affil_type == Affiliations:
            result = affiliation_data
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result

        if affil_type == Affiliation:
            result.add_affiliation(affiliation=affiliation_data, use_api=use_api)
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result

        if affil_type == pd.Series:
            affiliation = Affiliation()
            affiliation.add_series(affiliation_data)
            result.add_affiliation(affiliation=affiliation, use_api=use_api)
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result

        if affil_type == pd.DataFrame:
            result.import_crossref_result(affiliation_data, use_api=use_api) # type: ignore
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result

        if affil_type == dict:
            affiliation = Affiliation.from_dict(affiliation_data) # type: ignore
            result.add_affiliation(affiliation = affiliation, use_api=use_api) # type: ignore
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result


        if (affil_type == list) and (len(affiliation_data) > 0) and (type(affiliation_data[0]) == Affiliation):
            result = Affiliations()
            result.add_list(affiliation_data)
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
            return result
            

        if (affil_type == list) and (len(affiliation_data) > 0) and (type(affiliation_data[0]) == dict):

            for i in affiliation_data:
                affiliation = Affiliation.from_dict(i) # type: ignore
                result.add_affiliation(affiliation = affiliation, use_api=use_api) # type: ignore
            
            if drop_empty_rows == True:
                result.drop_empty_rows()
            if drop_duplicates == True:
                result.remove_duplicates(drop_empty_rows=drop_empty_rows)
    
            return result


from ..datasets import all_personal_names, first_names, last_names, nltk_names, countries_all, country_names, cities_all, cities_en, language_names, languages_en, language_codes, stopwords, languages_all, countries_zh, countries_ar, countries_es, countries_hi, countries_pt, countries_ru, countries_fr, nltk_words_list, nltk_webtext_words

from typing import List, Dict, Tuple
import copy
import pandas as pd

def extract_names(words: list, source: str = 'all_personal_names'):
    
    """
    Takes a list of words and returns all words that appear in a pre-defined list of names.
    
    Parameters
    ----------
    words : list, set, or tuple
        iterable of strings to analyse.
    source : str
        name of corpus of names to use for comparison.
    
    Returns
    -------
    result : list
        extracted names.
    
    Notes
    -----
        * 'all_personal_names' corpus includes first names and last names from a large number of languages. It features >1,000,000 names.
    """
    
    # Formatting words list and removing duplicates
    words = {i.strip('.').strip().lower() for i in words}
    
    
    
    # Retrieving specified name corpus and formatting it
    names_set = None
    
    if source == 'all_personal_names':
   
        global all_personal_names
        names_set = pd.Series(all_personal_names).str.lower().str.strip('.').to_list()
    
    if source == 'first_names':
        
        global first_names
        names_set = pd.Series(first_names).str.lower().str.strip('.').to_list()
    
    if source == 'last_names':
        
        global last_names
        names_set = set(pd.Series(last_names).str.lower().str.strip('.').to_list())
    
    if source == 'nltk_names':
        
        global nltk_names
        names_set = set(pd.Series(nltk_names).str.lower().str.strip('.').str.strip().to_list())
    
    if names_set == None:
        names_set = set(source)
    
    # Identifying intersection of names list and words list. Converting to a list.
    extracted_names = list(words.intersection(names_set))
    
    return extracted_names
    

def extract_countries(words, language = 'all'):
    
    """
    Takes a list of words and returns all words that appear in a pre-defined list of countries.
    
    Parameters
    ----------
    words : list, set, or tuple
        iterable of strings to analyse.
    language : str
        language of corpus of countries to use for comparison.
    
    Returns
    -------
    result : list
        extracted country names.
    
    Notes
    -----
        * 'all' corpus includes country names from a large number of languages.
    """
    
    # Formatting words list and removing duplicates
    words = {i.strip('.').strip().lower() for i in words}
    
    # Retrieving all country names corpus and formatting it
    if language == 'all':
        
        global countries_all
        countries_set = set(pd.Series(countries_all).str.lower().str.strip('.').to_list())
    
    # Retrieving country names corpus if a language is specified and formatting it
    else:
        global country_names
        countries = country_names[language]
        countries_set = set(pd.Series(countries).str.lower().str.strip('.').to_list())
    
    # Identifying intersection of country names corpus and words list. Converting to a list.
    extracted_countries = list(words.intersection(countries_set))
    
    return extracted_countries
    

def extract_cities(words, language = 'all'):
    
    """
    Takes a list of words and returns all words that appear in a pre-defined list of cities.
    
    Parameters
    ----------
    words : list, set, or tuple
        iterable of strings to analyse.
    language : str
        language of corpus of city names to use for comparison.
    
    Returns
    -------
    result : list
        extracted city names.
    
    Notes
    -----
        * 'all' corpus includes city names from a large number of languages.
    """
    
    # Formatting words list and removing duplicates
    words = {i.strip('.').strip().lower() for i in words}
    
    # Retrieving all city names corpus and formatting it
    if language == 'all':
        
        global cities_all
        cities_set = set(pd.Series(cities_all).str.lower().str.strip('.').to_list())
    
    # Retrieving city names corpus if the language is specified as English and formatting it
    if (language.lower() == 'en') or (language.lower() == 'english'):
        
        global cities_en
        cities = list(cities_en.keys())
        cities_set = set(pd.Series(cities).str.lower().str.strip('.').to_list())
    
    # Identifying intersection of city names corpus and words list. Converting to a list.
    extracted_cities = list(words.intersection(cities_set))
    
    return extracted_cities
    

def extract_languages(words, language = 'all'):
    
    """
    Takes a list of words and returns all words that appear in a pre-defined list of language names.
    
    Parameters
    ----------
    words : list, set, or tuple
        iterable of strings to analyse.
    language : str
        language of corpus of language names to use for comparison.
    
    Returns
    -------
    result : list
        extracted language names.
    
    Notes
    -----
        * 'all' corpus includes language names from a large number of languages.
    """
    
    # Formatting words list and removing duplicates
    words = {i.strip('.').strip().lower() for i in words}
    
    # Retrieving all language names corpus and formatting it
    if language == 'all':
        
        global languages_all
        langs_set = set(pd.Series(languages_all).str.lower().str.strip('.').to_list())
    
    # Retrieving language names corpus if the language is specified as English and formatting it
    if (language.lower() == 'en') or (language.lower() == 'english'):
        
        global languages_en
        langs = list(languages_en.values())
        langs_set = set(pd.Series(langs).str.lower().str.strip('.').to_list())
    
    # Identifying intersection of city names corpus and words list. Converting to a list.
    extracted_langs = list(words.intersection(langs_set))
    
    return extracted_langs
    



from ..utils.basics import Iterator
from datetime import datetime
from pathlib import Path

class Properties:
    
    """
    This is a class for properties to be assigned to Reviews.
    
    Parameters
    ----------
    review_name : str 
        name of the Review in the environment.
    file_location : str 
        directory address for source file.
    file_type : str 
        file type for source file.
        
    Attributes
    ----------
    review_name : str
        name of Review in the environment.
    obj_path : str
        path to object in the environment.
    created_at : str
        date and time created.
    last_changed_at : str
        date and time the object was last edited.
    obj_size : float
        size of the object in memory in bytes.
    """
    
    def __init__(self, review_name = None, file_location = None):
        
        """
        Initialises Properties instance.
        
        Parameters
        ----------
        parent_obj_path : str 
            name of parent object if object is an attribute of another object.
        size : int 
            size of object in memory.
        """
        if review_name == None:
            review_name = 'Review'

        self.review_name = review_name
        self.created_at = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        self.last_changed_at = self.created_at
        self.last_backup = None
        self.file_location = file_location

        if file_location is not None:
            file_type = Path(file_location).suffix
            self.file_type = file_type
        else:
            self.file_type = None
    
    def __iter__(self):
        
        """
        Function to make Properties objects iterable.
        """
        
        return Iterator(self)
    
    def __repr__(self):
        
        """
        Defines how Properties objects are represented in string form.
        """
        
        self_dict = self.to_dict()
        output = '\n'
        for key in self_dict.keys():
            prop = self_dict[key]
            output = output + key + ': ' + str(prop) + '\n'
        
        return output
    
    def to_list(self):
        
        """
        Returns Properties object as a list.
        """
        
        return [i for i in self]

    def to_dict(self):
        
        """
        Returns Properties object as a dictionary.
        """
        
        output_dict = {}
        for index in self.__dict__.keys():
            output_dict[index] = self.__dict__[index]

        return output_dict
    
    
    def update_last_changed(self):
        
        """
        Updates the last_changed attribute to the current date and time.
        """
        
        self.last_changed_at = datetime.now().strftime("%d/%m/%Y %H:%M:%S")

    def update_file_type(self):

        """
        Updates the file_type attribute to the current file_location string's suffix.
        """

        file_path = self.file_location

        if file_path is not None:
            self.file_type = Path(file_path).suffix
        else:
            self.file_type = None



"""
==========================
Academic Review Tool (ART)
==========================

The Academic Review Tool (ART) is a package for performing academic reviews and bibliometric analyses in Python. 
It offers capabilities for discovering, retrieving, and analysing academic literature at scale. 
ART accesses records from Crossref, Web of Science, Scopus, Orcid, and more.

ART's functionalities include:
* Searching for works using keywords, dates, authors, funders, and other information.
* Searching for authors and their works.
* Searching for funders and their funded works.
* Looking up DOIs, ISBNs, ISSNs, ORCID IDs, URLs, and other unique identifiers.
* Scraping academic repositories and websites.
* Citation and weblink crawling.
* Citation and coauthorship analysis.
* Generating networks representing:
    ** Citations and references
    ** Coauthors
    ** Cofunders
    ** And more...

ART uses the following APIs:
* Crossref
* Web of Science (Starter API)
* Scopus
* ORCID
* Geopy / Nominatim
"""

from .importers.crossref import lookup_doi, lookup_dois, lookup_journal, lookup_journals, search_journals, get_journal_entries, search_journal_entries, lookup_funder, lookup_funders, search_funders, get_funder_works, search_funder_works
from .importers.crossref import search_works as search_crossref
# from .importers.wos import search as search_wos
from .importers.scopus import search as search_scopus, lookup as lookup_scopus
from .importers.orcid import lookup_orcid, search as search_orcid
from .importers.search import search as api_search
# from .importers import pdf, orcid, crossref, scopus, jstor, wos
from .classes import Results, References, Author, Authors, Funder, Funders, Affiliation, Affiliations, Review
from .classes.networks import Network, Networks
from .classes.citation_crawler import academic_scraper as scrape

import pickle

def open_file(file_address: str = 'request_input'): # type: ignore

        if file_address == 'request_input':
            file_address = input('File address: ')
        
        if (file_address.endswith('.txt')) or (file_address.endswith('.review')):
            with open(file_address, 'rb') as f:
                review = pickle.load(f)

        return review


from ..utils.basics import results_cols

import pybtex # type: ignore
import pandas as pd
import numpy as np

def import_bibtex(file_path = 'request_input'):

    if file_path == 'request_input':
        file_path = input('File path: ')

    bib = pybtex.database.parse_file(file_path)

    df = pd.DataFrame(columns=results_cols, dtype=object)

    bib_keys = list(bib.entries.keys())

    index = 0
    for k in bib_keys:

        entry = bib.entries.get(k)

        fields = dict(entry.fields)
        field_keys = fields.keys()

        df.loc[index, 'type'] = entry.type

        if 'title' in field_keys:
            df.loc[index, 'title'] = fields['title']
        
        if 'year' in field_keys:
            df.loc[index, 'date'] = fields['year']
        
        if 'url' in field_keys:
            df.loc[index, 'link'] = fields['url']
        
        if 'doi' in field_keys:
            df.loc[index, 'doi'] = fields['doi']

        if 'language' in field_keys:
            df.loc[index, 'language'] = fields['language']
        
        if 'journal' in field_keys:
            df.loc[index, 'source'] = fields['journal']
        
        if 'book' in field_keys:
            df.loc[index, 'source'] = fields['book']
        
        if 'booktitle' in field_keys:
            df.loc[index, 'source'] = fields['booktitle']
        
        if 'publisher' in field_keys:
            df.loc[index, 'publisher'] = fields['publisher']
        
        if 'pmid' in field_keys:
            df.loc[index, 'pubmed_id'] = fields['pmid']

        if 'issn' in field_keys:
            df.loc[index, 'issn'] = fields['issn']
        
        if 'isbn' in field_keys:
            df.loc[index, 'isbn'] = fields['isbn']
        
        if 'keywords' in field_keys:
            df.at[index, 'keywords'] = fields['keywords'].split(',')
        
        persons = dict(entry.persons)
        persons_keys = persons.keys()

        if 'author' in persons_keys:
            auths = persons['author']
            auths_list = []

            for a in auths:
                first = ''.join(a.first_names)
                middle = ' '.join(a.middle_names)
                last = ' '.join(a.last_names)

                full_name = first + ' ' + middle + ' ' + last
                full_name = full_name.strip()
                auths_list.append(full_name)

            df.at[index, 'authors'] = auths_list
            df.at[index, 'authors_data'] = auths_list
        
        index += 1


    return df


import typing_extensions

from clarivate.wos_starter.client.paths import PathValues
from clarivate.wos_starter.client.apis.paths.journals import Journals
from clarivate.wos_starter.client.apis.paths.journals_id import JournalsId
from clarivate.wos_starter.client.apis.paths.documents import Documents
from clarivate.wos_starter.client.apis.paths.documents_uid import DocumentsUid

PathToApi = typing_extensions.TypedDict(
    'PathToApi',
    {
        PathValues.JOURNALS: Journals,
        PathValues.JOURNALS_ID: JournalsId,
        PathValues.DOCUMENTS: Documents,
        PathValues.DOCUMENTS_UID: DocumentsUid,
    }
)

path_to_api = PathToApi(
    {
        PathValues.JOURNALS: Journals,
        PathValues.JOURNALS_ID: JournalsId,
        PathValues.DOCUMENTS: Documents,
        PathValues.DOCUMENTS_UID: DocumentsUid,
    }
)



"""Functions to interface with the ORCID API"""

from ..utils.basics import results_cols
from ..utils.cleaners import is_int
from ..internet.webanalysis import is_url

from time import sleep

from requests import RequestException
from pathlib import Path

import pyorcid # type: ignore
from pyorcid import OrcidAuthentication, Orcid, OrcidSearch # type: ignore

import pandas as pd
import numpy as np

client_id = 'APP-105SVMUPF0B5U042'
client_secret = 'c7455d6f-f546-4d01-af46-2c314a4a8f46'
orcid_auth = OrcidAuthentication(client_id=client_id, client_secret=client_secret)

public_access_token = orcid_auth.get_public_access_token()

def lookup_orcid(orcid_id = 'request_input'):

    if orcid_id == 'request_input':
        orcid_id = input('ORCID ID: ')
    
    orcid_id = str(orcid_id).replace('https://', '').replace('http://', '').replace('orcid.org/', '')

    try:
        orcid = pyorcid.OrcidScrapper(orcid_id=orcid_id)

        result = orcid.record_summary()
        df = pd.DataFrame.from_dict(result, orient='index').T
        df.columns = df.columns.str.lower()
    except:
        print(f'ORCiD lookup for {orcid_id} failed.')
        df = pd.DataFrame(columns = ['orcid id', 'last modified', 'name', 'family name', 'credit name', 'other names', 'biography', 'emails', 'research tags (keywords)', 'employment', 'distinctions', 'fundings', 'works'],
                          dtype=object)

    return df

def get_author(orcid_id = 'request_input'):

    if orcid_id == 'request_input':
        orcid_id = input('ORCID ID: ')
    
    global public_access_token
    orcid = Orcid(orcid_id=orcid_id, orcid_access_token=public_access_token, state = "public")

    return orcid

def get_author_works(orcid_id = 'request_input', output = 'dataframe'):

    if orcid_id == 'request_input':
        orcid_id = input('ORCID ID: ')
    
    orcid = get_author(orcid_id=orcid_id)
    works_tuple = orcid.works()
    if len(works_tuple) > 0:
        works_list = works_tuple[0]
    else:
        works_list = []

    if (output == tuple) or (output == 'tuple'):
        return works_tuple

    if (output == list) or (output == 'list'):
        return works_list
    
    if output == 'dataframe':
        return pd.DataFrame(works_list)

def search(query: str = 'request_input', start: int = 0, limit: int = 1000, output: str = 'dataframe'):

    if query == 'request_input':
        query = input('Search query: ')

    global public_access_token
    
    orcidSearch = OrcidSearch(orcid_access_token=public_access_token)
    results = orcidSearch.search(query=query, start=start, rows=limit)

    error_msg = ''
    if 'response-code' in results.keys():
        error_msg = f'{error_msg}Response code: {results["response-code"]}.'
    
    if 'error-code' in results.keys():
        error_msg = f'{error_msg} Error code: {results["error-code"]}.'
    
    if 'developer-message' in results.keys():
        error_msg = f'{error_msg} {results["developer-message"]}.'

    if 'more-info' in results.keys():
        error_msg = f'{error_msg} For more info, see: {results["more-info"]}.'

    if len(error_msg) > 0:
        raise ValueError(f'{error_msg}')

    if 'num-found' in results.keys():
        num_found = results['num-found']
        print(f'{num_found} results found') # type: ignore

    if 'expanded-result' in results.keys():
        results_list = results['expanded-result']
    else:
        results_list = []

    if (output == dict) or (output.lower().strip() == 'dict'):
        return results

    if (output == list) or (output.lower().strip() == 'list'):
        return results_list
    
    if (output == pd.DataFrame) or (output.lower().strip() == 'dataframe'):
        return pd.DataFrame(results_list)



    


def save_summary(self: Orcid, file_name: str = 'request_input', file_path: str = 'request_input'):

    if file_name == 'request_input':
        file_name = input('File name: ')
    
    file_name = file_name.strip('.md')
    
    if file_path == 'request_input':
        file_path = input('File path: ')
    
    path_obj = Path(file_path)

    if path_obj.exists() == False:
        raise ValueError('File path {file_path} does not exist')

    new_addr = file_path + '/' + file_name + '.md'

    self.generate_markdown_file(output_file=new_addr)

Orcid.save_summary = save_summary


from ..utils.basics import results_cols

from igraph import Graph # type: ignore
from networkx.classes import Graph as NetworkX_Undir, DiGraph as NetworkX_Dir, MultiGraph as NetworkX_Multi # type: ignore
import numpy as np
import pandas as pd

import itertools

def generate_urls_network(urls_dict: dict) -> Graph:
        
        """
        Returns a directed network representing if/how items embed hyperlinks to one another.
        
        Parameters
        ----------
        urls_dict : dict
            a dictionary with the following structure:
                * keys: URLs
                * values: links associated with the URLs
        """

        
        # Creating list of all URLs in source dictionary
        all_urls = []
        for i in urls_dict.keys():
            all_urls.append(i)
            all_urls = all_urls + urls_dict[i]
        
        # Removing repeats
        all_urls = set(all_urls)
        
        # Removing None and empty string URLs
        all_urls = [i for i in all_urls if ((i != None) and (i != ''))]
        
        # Initialising network
        g = Graph(n=len(all_urls), directed=True, vertex_attrs={'name': all_urls})
        
        
        # Adding edges by iterating through vertices and retrieving links associated
        for vertex in g.vs:
            
            # Getting vertex url
            url = vertex['name']
            
            # Ignoring None and empty string URL addresses
            if (url != None) and (len(url) > 0):
                
                # Retrieving vertex index
                v_index = vertex.index
                
                # Retrieving vertex links
                if url in urls_dict.keys():
                    v_links = urls_dict[url]
                else:
                    v_links = []
                
                # If the vertex has links associated, finds vertices those links direct to
                if len(v_links) > 0:
                    
                    for link in v_links:
                        end_index = g.vs.find(name = link).index
                        
                        # Adding edges between linked vertices
                        Graph.add_edges(g, 
                                        [(v_index, end_index)], 
                                        attributes={
                                           'name': f'{url} -> {link}'})
                    
        return g

def colinks_out(links_network: Graph) -> pd.DataFrame:
    
    """
    Conducts out-bound colink analysis. Returns a pandas.DataFrame.
    
    Notes
    -----
    Is able to take igraph.Graph, Network, and NetworkX objects.
    """
    
    # Converting NetworkX objects to igraph objects
    if (
            (type(links_network) == NetworkX_Undir)
            or (type(links_network) == NetworkX_Dir)
            or (type(links_network) == NetworkX_Multi)
        ):
            links_network = Graph.from_networkx(links_network)
    
    # Retrieving URLs list
    urls = links_network.vs['name']
    
    # Initialising dictionary to store tuples of co-linked URLs
    colink_tuples_dict = {}
    
    # Iterating through vertices to identify pairs of co-linked URLs
    for v in links_network.vs:
        
        # Retrieving vertex URL
        url = v['name']
        
        # Creating list of URLS linked by selected URL 
        children = v.successors()
        names = [v['name'] for v in children]
        
        # Creating list of all combinations of linked URLs
        # Stored as a list of tuples
        tuples = list(itertools.combinations(names, 2))
        tuples = [tuple(set(i)) for i in tuples]
        
        # Iterating through linked URLs combinations and adding to dictionary of colinks
        for pair in tuples:
            
            # Adding to dictionary of colinks
            if pair not in colink_tuples_dict.keys():
                colink_tuples_dict[pair] = {'Co-link to': [], 'Frequency': 0}

            colink_tuples_dict[pair]['Co-link to'].append(url)
            
            # Adding combination to count
            colink_tuples_dict[pair]['Frequency'] += 1

    # Creating and formatting output dataframe
    df = pd.DataFrame.from_dict(colink_tuples_dict).T
    df.index.name = 'Names'
    df = df.sort_values('Frequency', ascending=False)
    
    return df

def colinks_in(links_network: Graph) -> pd.DataFrame:
    
    """
    Conducts in-bound colink analysis. Returns a pandas.DataFrame.
    
    Notes
    -----
    Is able to take igraph.Graph, Network, and NetworkX objects.
    """
    
    # Converting NetworkX objects to igraph objects
    if (
            (type(links_network) == NetworkX_Undir)
            or (type(links_network) == NetworkX_Dir)
            or (type(links_network) == NetworkX_Multi)
        ):
            links_network = Graph.from_networkx(links_network)
    
    # Retrieving URLs list
    urls = links_network.vs['name']
    
    # Initialising dictionary to store tuples of co-linking URLs
    coupling_dict = {}
    
    # Iterating through vertices to identify pairs of co-linking URLs
    for v in links_network.vs:
        
        # Retrieving vertex URL
        url = v['name']
        
        # Creating list of URLS linking to selected URL 
        parents = v.predecessors()
        names = [v['name'] for v in parents]
        
        # Creating list of all combinations of linking URLs
        # Stored as a list of tuples
        tuples = list(itertools.combinations(names, 2))
        tuples = [tuple(set(i)) for i in tuples]
        
         # Adding to dictionary of colinking URLs
        if url not in coupling_dict.keys():
            coupling_dict[url] = {'Co-linked by': [], 'Frequency': 0}
        
        coupling_dict[url]['Co-linked by'] = list(set(coupling_dict[url]['Co-linked by'] + tuples))
        
        # Counting and storing number of times selected URL is co-linked to
        coupling_dict[url]['Frequency'] = len(coupling_dict[url]['Co-linked by'])
    
    # Creating and formatting output dataframe
    df = pd.DataFrame.from_dict(coupling_dict).T
    df.index.name = 'Name'
    df = df.sort_values('Frequency', ascending=False)
    
    return df



def generate_coauthors_network(coauthors_dict: dict) -> Graph:
        
        """
        Returns an undirected network representing how authors co-publish with one another.
        
        Parameters
        ----------
        coauthors_dict : dict
            a dictionary with the following structure:
                * keys: author IDs
                * values: Pandas DataFrames containing details on co-authors.
        
        Returns
        -------
        g : Graph
            an iGraph Graph object representing the co-authorship network.
        """

        
        # Creating list of all unique author_ids in source dictionary
        all_auths = list(set(coauthors_dict.keys()))
        
        
        # Initialising network
        g = Graph(n=len(all_auths), directed=False, vertex_attrs={'name': all_auths})
        
        
        # Adding edges by iterating through vertices and retrieving links associated
        for auth_id in all_auths:
            
            # Ignoring None and empty string vertex IDs
            if (auth_id != None) and (len(auth_id) > 0):
                
                # Adding vertex if none in graph
                if auth_id not in g.vs['name']:
                    g.add_vertex(name=auth_id)

                # Getting vertex
                vertex = g.vs.find(name=auth_id)
            
                # Retrieving vertex index
                v_index = vertex.index
                
                # Retrieving co-author ids
                v_edges = coauthors_dict[auth_id]['author_id'].to_list()
                
                # If the vertex has links associated, finds vertices those links direct to
                if len(v_edges) > 0:
                    
                    for author in v_edges:

                        if author not in g.vs['name']:
                            g.add_vertex(name=author)

                        end_index = g.vs.find(name = author).index
                        df_index = coauthors_dict[auth_id][coauthors_dict[auth_id]['author_id'] == author].index.to_list()[0]
                        weight = coauthors_dict[auth_id].loc[df_index, 'frequency']
                        
                        # Adding edges between linked vertices
                        Graph.add_edges(g, 
                                        [(v_index, end_index)], 
                                        attributes={
                                           'name': f'{auth_id} <-> {author}',
                                           'weight': weight
                                           })
        
        g = g.simplify(combine_edges='first')

        return g

def generate_funders_network(funders_dict: dict) -> Graph:
        
        """
        Returns an undirected network representing how funders co-publish with one another.
        
        Parameters
        ----------
        funders_dict : dict
            a dictionary with the following structure:
                * keys: funder IDs
                * values: Pandas DataFrames containing details on co-funders.
        
        Returns
        -------
        g : Graph
            an iGraph Graph object representing the co-funder network.
        """

        
        # Creating list of all unique funder_ids in source dictionary
        all_funders = list(set(funders_dict.keys()))
        
        
        # Initialising network
        g = Graph(n=len(all_funders), directed=False, vertex_attrs={'name': all_funders})
        
        
        # Adding edges by iterating through vertices and retrieving links associated
        for funder_id in all_funders:
            
            
            # Ignoring None and empty string vertex IDs
            if (funder_id != None) and (len(funder_id) > 0):
                
                if funder_id not in g.vs['name']:
                    g.add_vertex(name=funder_id)

                # Getting vertex id
                vertex = g.vs.find(name = funder_id)

                # Retrieving vertex index
                v_index = vertex.index
                
                # Retrieving co-funder ids
                v_edges = funders_dict[funder_id]['funder_id'].to_list()
                
                # If the vertex has links associated, finds vertices those links direct to
                if len(v_edges) > 0:
                    
                    for f in v_edges:

                        if f not in g.vs['name']:
                            g.add_vertex(name=f)

                        end_index = g.vs.find(name = f).index
                        df_index = funders_dict[funder_id][funders_dict[funder_id]['funder_id'] == f].index.to_list()[0]
                        weight = funders_dict[funder_id].loc[df_index, 'frequency']
                        
                        # Adding edges between linked vertices
                        Graph.add_edges(g, 
                                        [(v_index, end_index)], 
                                        attributes={
                                           'name': f'{funder_id} <-> {f}',
                                           'weight': weight
                                           })
        
        g = g.simplify(combine_edges='first')

        return g

def generate_citations_network(citations_dict: dict) -> Graph:
        
        """
        Returns an undirected network representing how publications cite one another.
        
        Parameters
        ----------
        citations_dict : dict
            a dictionary with the following structure:
                * keys: work IDs
                * values: Pandas DataFrames containing details on cited works.
        
        Returns
        -------
        g : Graph
            an iGraph Graph object representing the citation network.
        """

        
        # Creating list of all unique author_ids in source dictionary
        all_ids = list(set(citations_dict.keys()))
        
        
        # Initialising network
        g = Graph(n=len(all_ids), directed=True, vertex_attrs={'name': all_ids})
        
        
        # Adding edges by iterating through vertices and retrieving links associated
        for work_id in all_ids:
            
            # Ignoring None and empty string vertex IDs
            if (work_id != None) and (len(work_id) > 0):
                
                if work_id not in g.vs['name']:
                    g.add_vertex(name=work_id)

                # Getting vertex object
                vertex = g.vs.find(name = work_id)
                
                # Retrieving vertex index
                v_index = vertex.index
                
                refs_obj = citations_dict[work_id]
                if (refs_obj is None) or (type(refs_obj) == float ) or (refs_obj is np.nan) or (type(refs_obj) == str):
                     continue
                
                if (refs_obj is not None) and (type(refs_obj) != float ) and (refs_obj is not np.nan) and (type(refs_obj) != str):
                    if 'update_work_ids' in refs_obj.__dir__():
                        refs_obj.update_work_ids()

                    if 'copy' in refs_obj.__dir__():
                        df = refs_obj.copy(deep=True)
                        
                    else:
                         df = refs_obj
                    
                    if 'astype' in df.__dir__():
                            df = df.astype(str)

                else:
                    df = pd.DataFrame(columns=results_cols, dtype=object)

                if type(df) == float:
                     df = pd.DataFrame(columns=results_cols, dtype=object)

                # Retrieving citations work ids
                df_index = df.index.to_list()
                
                # If the vertex has links associated, finds vertices those links direct to
                if len(df_index) > 0:
                    
                    for i in df_index:

                        citation_data = refs_obj.loc[i] # type: ignore
                        citation = citation_data['work_id']

                        if citation not in g.vs['name']:
                             g.add_vertex(name=citation)

                        end_index = g.vs.find(name = citation).index

                        for i in citation_data.index.to_list():
                             data = citation_data[i]
                             g.vs[end_index][i] = data
                        
                        # Adding edges between linked vertices
                        Graph.add_edges(g, 
                                        [(v_index, end_index)], 
                                        attributes={
                                           'name': f'{work_id} -> {citation}'
                                           })
        
        g = g.simplify(combine_edges='first')

        return g

def generate_author_works_network(author_works_dict: dict) -> Graph:

    """
        Returns an undirected bipartite network representing the relationships between authors and publications.
        
        Parameters
        ----------
        author_works_dict : dict
            a dictionary with the following structure:
                * keys: work IDs
                * values: Pandas DataFrames containing details on authors.
        
        Returns
        -------
        g : Graph
            an iGraph Graph object.
    """

    work_ids = author_works_dict.keys()
    work_ids = [i for i in work_ids if (i != None) and (len(i) > 0)]
    works_len = len(work_ids)
    init_types = [False]*works_len

    g = Graph.Bipartite(types=init_types, edges=[], directed=False)
    g.vs['name'] = work_ids
    g.vs['category'] = ['publication']*works_len

    for work_id in work_ids:

        # Getting vertex object
        vertex = g.vs.find(name = work_id)

        v_index = vertex.index

        data = author_works_dict[work_id]

        if type(data) == dict:
             data = pd.DataFrame.from_dict(data, orient='index').T
        
        if type(data) != pd.DataFrame:
             continue
        
        if 'author_id' not in data.columns:
            continue
            
        auth_ids = data['author_id'].to_list()
        auth_ids = [i.split('#')[0].strip() for i in auth_ids if type(i) == str]

        for a in auth_ids:

            if a not in g.vs['name']:
                g.add_vertex(name=a, type=True)
                
            author_vertex = g.vs.find(name = a)
            author_index = author_vertex.index
            author_vertex['category'] = 'author'

            # Adding edges between linked vertices
            Graph.add_edges(g, 
                                        [(v_index, author_index)], 
                                        attributes={
                                           'name': f'{work_id} <-> {a}'
                                           })
    
    g = g.simplify(combine_edges='first')

    return g

def generate_author_affils_network(author_affils_dict: dict) -> Graph:

    """
        Returns an undirected bipartite network representing the relationships between authors and their affiliated organisations.
        
        Parameters
        ----------
        author_affils_dict : dict
            a dictionary with the following structure:
                * keys: work IDs
                * values: Pandas DataFrames containing details on authors.
        
        Returns
        -------
        g : Graph
            an iGraph Graph object.
    """

    author_ids = author_affils_dict.keys()
    author_ids_stripped = [i.split('#')[0].strip() for i in author_ids if ((type(i) == str) and (len(i) > 0))]
    auths_len = len(author_ids_stripped)
    init_types = [False]*auths_len

    g = Graph.Bipartite(types=init_types, edges=[], directed=False)
    g.vs['name'] = author_ids_stripped
    g.vs['category'] = ['author']*auths_len

    for auth_id in author_ids:
        
        auth_id_stripped = auth_id.split('#')[0].strip()

        # Getting vertex object
        vertex = g.vs.find(name = auth_id_stripped)

        v_index = vertex.index

        data = author_affils_dict[auth_id]
        
        if type(data) == dict:
             data = pd.DataFrame.from_dict(data, orient='index').T
        
        if type(data) != pd.DataFrame:
             continue

        if 'affiliation_id' not in data.columns:
            continue
            
        affil_ids = data['affiliation_id'].to_list()
        affil_ids = [i.split('#')[0].strip() for i in affil_ids if (type(i) == str and len(i) > 0)]

        for a in affil_ids:

            if a not in g.vs['name']:
                g.add_vertex(name=a, type=True)
                
            affil_vertex = g.vs.find(name = a)
            affil_index = affil_vertex.index
            affil_vertex['category'] = 'affiliation'

            # Adding edges between linked vertices
            Graph.add_edges(g, 
                                        [(v_index, affil_index)], 
                                        attributes={
                                           'name': f'{auth_id_stripped} <-> {a}'
                                           })
    
    g = g.simplify(combine_edges='first')

    return g

def generate_funder_works_network(funder_works_dict: dict) -> Graph:

    """
        Returns an undirected bipartite network representing the relationships between funders and publications.
        
        Parameters
        ----------
        author_works_dict : dict
            a dictionary with the following structure:
                * keys: work IDs
                * values: Pandas DataFrames containing details on funders.
        
        Returns
        -------
        g : Graph
            an iGraph Graph object.
    """

    work_ids = funder_works_dict.keys()
    work_ids = [i for i in work_ids if (i != None) and (len(i) > 0)]
    works_len = len(work_ids)
    init_types = [False]*works_len

    g = Graph.Bipartite(types=init_types, edges=[], directed=False)
    g.vs['name'] = work_ids
    g.vs['category'] = ['publication']*works_len

    for work_id in work_ids:

        # Getting vertex object
        vertex = g.vs.find(name = work_id)

        v_index = vertex.index

        data = funder_works_dict[work_id]

        if type(data) == dict:
             data = pd.DataFrame.from_dict(data, orient='index').T
        
        if type(data) != pd.DataFrame:
             continue

        if 'funder_id' not in data.columns:
            continue
            
        f_ids = data['funder_id'].to_list()
        f_ids = [i.split('#')[0].strip() for i in f_ids if type(i) == str]

        for f in f_ids:

            if f not in g.vs['name']:
                g.add_vertex(name=f, type=True)
                
            funder_vertex = g.vs.find(name = f)
            funder_index = funder_vertex.index
            funder_vertex['category'] = 'funder'

            # Adding edges between linked vertices
            Graph.add_edges(g, 
                                        [(v_index, funder_index)], 
                                        attributes={
                                           'name': f'{work_id} <-> {f}'
                                           })
    
    g = g.simplify(combine_edges='first')

    return g

def cocitation_dict(citation_network) -> dict:
     
    # Converting NetworkX objects to igraph objects
    if (
            (type(citation_network) == NetworkX_Undir)
            or (type(citation_network) == NetworkX_Dir)
            or (type(citation_network) == NetworkX_Multi)
        ):
            citation_network = Graph.from_networkx(citation_network)
    
    # Retrieving URLs list
    ids = citation_network.vs['name']
    
    # Initialising dictionary to store tuples of co-linked URLs
    cocitation_tuples_dict = {}
    
    # Iterating through vertices to identify pairs of co-linked URLs
    for v in citation_network.vs:
        
        # Retrieving vertex URL
        work_id = v['name']
        
        # Creating list of URLS linked by selected URL 
        children = v.successors()
        child_ids = [v['name'] for v in children]
        
        # Creating list of all combinations of linked URLs
        # Stored as a list of tuples
        tuples = list(itertools.combinations(child_ids, 2))
        tuples = [tuple(set(i)) for i in tuples]
        
        # Iterating through linked URLs combinations and adding to dictionary of colinks
        for pair in tuples:
            
            # Adding to dictionary of colinks
            if pair not in cocitation_tuples_dict.keys():
                cocitation_tuples_dict[pair] = {'Co-cited by': [], 'Frequency': 0}

            cocitation_tuples_dict[pair]['Co-cited by'].append(work_id)
            
            # Adding combination to count
            cocitation_tuples_dict[pair]['Frequency'] = len(set(cocitation_tuples_dict[pair]['Co-cited by']))
    
    return cocitation_tuples_dict

def generate_cocitation_network(citation_network):
    
    """
    Generates a co-citation network from a citation network.
    
    Notes
    -----
    Is able to take igraph.Graph, Network, and NetworkX objects.
    """
    
    
    cocitations = cocitation_dict(citation_network)
    v_len = len(citation_network.vs)
    v_attr_keys = citation_network.vs.attributes()

    v_attrs = {}
    for a in v_attr_keys:
        v_attrs[a] = citation_network.vs[a]
    
    
    cocitation_graph = Graph(n=v_len, directed=False, vertex_attrs=v_attrs, edge_attrs={'name':[], 'weight': [], 'cocited_by': []})

    for k in cocitations.keys():

        v1_id = k[0]
        v2_id = k[1]
        freq = cocitations[k]['Frequency']
        cocited_by = cocitations[k]['Co-cited by']

        if v1_id not in cocitation_graph.vs['name']:
              cocitation_graph.add_vertex(name=v1_id)
        
        if v2_id not in cocitation_graph.vs['name']:
              cocitation_graph.add_vertex(name=v2_id)
        
        v1 = cocitation_graph.vs.find(name=v1_id)
        v1_index = v1.index

        v2 = cocitation_graph.vs.find(name=v2_id)
        v2_index = v2.index

        if (cocitation_graph.are_connected(v1_index, v2_index) == False) and (cocitation_graph.are_connected(v2_index, v1_index) == False):
             Graph.add_edges(cocitation_graph, 
                                        [(v1_index, v2_index)], 
                                        attributes={
                                           'name': f'{v1_id} <-> {v2_id}',
                                           'weight': freq,
                                           'cocited_by': cocited_by
                                           })
        
        else:
             edgelist = list(cocitation_graph.es.select(_between= ([v1_index], [v2_index])))

             if len(edgelist) > 0:

                edge = edgelist[0]
                edge_index = edge.index

                old_cocited_by = cocitation_graph.es[edge_index]['cocited_by']
                new_cocited_by = old_cocited_by + cocited_by
                new_cocited_by = list(set(new_cocited_by))
                new_freq = len(new_cocited_by)

                cocitation_graph.es[edge_index]['cocited_by'] = new_cocited_by
                cocitation_graph.es[edge_index]['weight'] = new_freq
    
    # cocitation_graph = cocitation_graph.simplify()

    for v in citation_network.vs:

        work_id = v['name']
        if work_id not in cocitation_graph.vs['name']:
              cocitation_graph.add_vertex(name=work_id)
        
        attrs = citation_network.vs.attributes()
        cocit_v = cocitation_graph.vs.find(name=work_id)

        for a in attrs:
             cocit_v[a] = v[a]
             
    return cocitation_graph

def bibcoupling_dict(citation_network):
    
    """
    Generates a dictionary representing bibliometric coupling from a citation network.
    
    Notes
    -----
    Is able to take igraph.Graph, Network, and NetworkX objects.
    """
    
    # Converting NetworkX objects to igraph objects
    if (
            (type(citation_network) == NetworkX_Undir)
            or (type(citation_network) == NetworkX_Dir)
            or (type(citation_network) == NetworkX_Multi)
        ):
            citation_network = Graph.from_networkx(citation_network)
    
    # Initialising dictionary to store tuples of co-citing works
    bc_dict = {}
    
    # Iterating through vertices to identify pairs of co-citing works
    for v in citation_network.vs:
        
        # Retrieving work ID
        work_id = v['name']
        
        # Creating list of IDs citing the selected work 
        parents = v.predecessors()
        p_ids = [i['name'] for i in parents]
        
        # Creating list of all combinations of citing works
        # Stored as a list of tuples
        tuples = list(itertools.combinations(p_ids, 2))
        tuples = [tuple(set(i)) for i in tuples]

        for t in tuples:

            if t not in bc_dict.keys():
                bc_dict[t] = {'co-cite': [], 'frequency': 0}
            
            bc_dict[t]['co-cite'].append(work_id)
            bc_dict[t]['frequency'] = len(bc_dict[t]['co-cite'])
    
    
    
    return bc_dict


def generate_bibcoupling_network(citation_network):
    
    """
    Generates a bibliometric coupling network from a citation network.
    
    Notes
    -----
    Is able to take igraph.Graph, Network, and NetworkX objects.
    """
    
    bc_dict = bibcoupling_dict(citation_network)
    
    v_count = len(citation_network.vs)
    v_attr_keys = citation_network.vs.attributes()

    v_attrs = {}
    for a in v_attr_keys:
        v_attrs[a] = citation_network.vs[a]
    
    g = Graph(n=v_count, directed=False, vertex_attrs=v_attrs)

    for k in bc_dict.keys():
    
        v1_id = k[0]
        v2_id = k[1]
        freq = bc_dict[k]['frequency']
        cocite = bc_dict[k]['co-cite']

        if v1_id not in g.vs['name']:
              g.add_vertex(name=v1_id)
        
        if v2_id not in g.vs['name']:
              g.add_vertex(name=v2_id)
        
        v1 = g.vs.find(name=v1_id)
        v1_index = v1.index

        v2 = g.vs.find(name=v2_id)
        v2_index = v2.index

        if (g.are_connected(v1_index, v2_index) == False) and (g.are_connected(v2_index, v1_index) == False):
             Graph.add_edges(g, 
                                        [(v1_index, v2_index)], 
                                        attributes={
                                           'name': f'{v1_id} <-> {v2_id}',
                                           'weight': freq,
                                           'both_cite': cocite
                                           })
        
        else:
             edgelist = list(g.es.select(_between= ([v1_index], [v2_index])))

             if len(edgelist) > 0:

                edge = edgelist[0]
                edge_index = edge.index

                old_cociting = g.es[edge_index]['both_cite']
                new_cociting = old_cociting + cocite
                new_cociting = list(set(new_cociting))
                new_freq = len(new_cociting)

                g.es[edge_index]['both_cite'] = new_cociting
                g.es[edge_index]['weight'] = new_freq
    
    
    
    return g




"""Variables and functions for interacting with the global environment."""

def get_var_name_str(variable):
    
    """
    Returns the string name of a variable as defined in the global or local environment.
    """
    
    # Searching global variables dictionary for key-value pairs that are identical to the item
    try:
        for key in globals().keys():
            if globals()[key] == variable:
                return key
    
    # Searching local variables dictionary for key-value pairs that are identical to the item
    except:
        try:
            for key in locals().keys():
                if locals()[key] == variable:
                    return key
        except:
            raise KeyError('No variable name found')


import typing_extensions

from clarivate.wos_starter.client.apis.tags import TagValues
from clarivate.wos_starter.client.apis.tags.documents_api import DocumentsApi
from clarivate.wos_starter.client.apis.tags.journals_api import JournalsApi

TagToApi = typing_extensions.TypedDict(
    'TagToApi',
    {
        TagValues.DOCUMENTS: DocumentsApi,
        TagValues.JOURNALS: JournalsApi,
    }
)

tag_to_api = TagToApi(
    {
        TagValues.DOCUMENTS: DocumentsApi,
        TagValues.JOURNALS: JournalsApi,
    }
)
